#!/usr/bin/perl -w
#
# SNMPAgent.pl
#
# Emulate an SNMP agent
#
# Andrew Findlay
# March 2003
# andrew.findlay@skills-1st.co.uk
#
# $Id: SNMPAgent.pm,v 1.15 2009/11/13 13:29:55 afindlay Exp $


#############################################################################
# 
# Copyright (c) 2003 Andrew Findlay and other contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
#############################################################################


########################################################################

=head1 Raddle SNMP Agent Emulator

The B<Net::Raddle::SNMPAgent> package provides complete control over the data
returned by an SNMP agent. This is useful when building emulated
networks for testing network management packages or teaching people to
use them.

In its simplest form, Raddle will replay static data captured from
a real SNMP agent (e.g. using the I<snmpwalk> command).

More advanced usage includes adding interfaces and entries in various
tables, to build up a complex emulated device.

=head2 Prerequisites

Raddle requires I<net-snmp> version 5.0.8 or later, built with the
Perl module enabled. If using 5.0.8, a patch must be applied to the
source code first. It is much better to use a later version:
version 5.2.1 has been tested and works OK.

The Perl modules Digest::MD5 and Carp are also required.

=head2 Example

To emulate a router called r1 using an I<snmpwalk> output file in
/usr/local/raddle/branch-network/r1.snmp, put
this in /usr/local/raddle/branch-network/r1.pl:

	use Net::Raddle::SNMPAgent;
	my $agent = Net::Raddle::SNMPAgent->new() or die "Could not initialise Raddle";
	$agent->ParseDataFile( '/usr/local/raddle/branch-network/r1.snmp', 0 );

This is the config file /usr/local/raddle/branch-network/r1.conf:

	# Listen on port 9501
	agentaddress    udp:localhost:9501,tcp:localhost:9501

	# Access control
	com2sec notConfigUser  default       public
	group   notConfigGroup v1            notConfigUser
	group   notConfigGroup v2c           notConfigUser
	view    iso included .iso
	access  notConfigGroup "" any noauth exact iso none none

	# Emulator script
	perl do "/usr/local/raddle/branch-network/r1.pl"

It is wise to test the emulator script before trying to run it with
snmpd, as the daemon seems to hide certain types of error report.
Do something like this:

	perl -w /usr/local/raddle/branch-network/r1.pl

Start I<snmpd> like this:

	snmpd -C -I vacm_conf -c /usr/local/raddle/branch-network/r1.conf

SNMPD will now respond on localhost:9501 using data from the C<r1.snmp> file.

=head2 Generating data files

Use I<snmpwalk> or I<snmpbulkwalk>:

       snmpwalk -v 1 -c <community> -OneU <device>

It is OK to miss the 'n' option from -O if you prefer to use symbolic names
but beware that net-snmp sometimes translates OIDs into unreadable text
that cannot be used.

=head2 Error Handling

Parameter errors and other serious problems are signalled using
C<confess> from the I<Carp> module. This in turn calls C<die>

=head2 Setup Procedures

=over

=cut

#
########################################################################
########################################################################


package Net::Raddle::SNMPAgent;

use strict;
use warnings;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
	IPToMAC
	Netmask
	CanonicalOID
);

use NetSNMP::OID (':all');
use NetSNMP::agent (':all');
use NetSNMP::ASN (':all');
use Digest::MD5 qw(md5_hex);
use Carp;
use Net::Raddle::TicksSince;
use Socket;

our $VERSION = 0.07;



########################################################################
########################################################################
# Main initialisation procedure

=item C<< Net::Raddle::SNMPAgent->new(params) >>

Initialises the SNMP emulation and links it to the main SNMP agent.

params is a key-value list, which can specify:

	debugging => 0,		# higher number for more debugging
	subagent => 0,		# Should this be an AgentX subagent?

If C<$subagent> is true, register as an AgentX subagent.
If C<$subagent> is false (e.g. zero), run as part of the main agent process.

e.g.:

	my $agent = Net::Raddle::SNMPAgent->new( debugging => 1 );

=cut

#
########################################################################
########################################################################

sub new () {
	my $invocant = shift;
	my $class = ref($invocant) || $invocant;
	my $self = {
		# Public variables
		#
		debugging => 0,
		subagent => 0,
		#
		# Private variables
		#
		# This is the object we get back from net-snmp
		snmp_agent => undef,
		#
		# Hash to store MIB data, indexed by numeric OID string
		mibdata => {},
		# List of OIDs used in mibdata, held in internal format
		oidlist => [],
		# Flag to indicate that @oidlist needs re-sorting
		oidlistDirty => 0,
		# Hash giving location of OIDs in @oidlist
		oidlistindex => {},
		#
		# Allow constructor to override defaults
		#
		@_,
	};


	# Different startup depending on whether we are running
	# as an AgentX subagent
	if ($self->{subagent}) {
		$self->{snmp_agent} = new NetSNMP::agent(
			'Name' => 'raddle', # reads raddle.conf???
			'AgentX' => 1);   # make us a subagent
		return 0 if (!$self->{snmp_agent});
		print STDERR "started Raddle as a subagent ($self->{snmp_agent})\n"
			if ($self->{debugging});
	}
	else {
		$self->{snmp_agent} = new NetSNMP::agent(
			Name => 'raddle', # reads raddle.conf???
			Ports => "udp:161,tcp:161",
			# dont_init_agent => 1,
			# dont_init_lib => 1,
		);
		return 0 if (!$self->{snmp_agent});
		print STDERR "started Raddle as an embedded part of the agent\n" if ($self->{debugging});
	}

	# We want to take over the whole tree but the module crashes when we try that :-(
	# Just go for .1.3 (.iso.org) to start with.
	my $regoid = new NetSNMP::OID('.1.3');
	print STDERR "registering at ",$regoid,"\n" if ($self->{debugging});

	# We register ourselves with the master agent we're embedded in.
	# We actually register a closure so that we can include the current $self data
	# which is needed by the handler
	#
	# We ignore the return code here as it does not match the manpage
	#
	my $res = $self->{snmp_agent}->register('raddle',$regoid, sub {my_snmp_handler($self,@_)} );

	print STDERR "registered OK: $res\n" if ($self->{debugging});

	# All done, so return self as an object
	return bless $self, $class;
}

######################################################################
######################################################################
#

=item C<< $agent->AgentXMainLoop() >>

If running as an AgentX subagent, this procedure must be called after
all other setup is complete. AgentXMainLoop only returns if the
process receives an INT or QUIT signal.

=cut

######################################################################
######################################################################
sub AgentXMainLoop ($) {
	my $self = shift;

	# We need to perform a loop here waiting for snmp requests.  We
	# aren't doing anything else here, but we could.
	while(1) {
		$self->{snmp_agent}->agent_check_and_process(1);  # 1 = block
		print STDERR "mainloop excercised\n" if ($self->{debugging});
	}
}



########################################################################
#
# Procedures to parse a file generated by commands like:
#	snmpbulkwalk -c <community> -OneU -v 2c <device>
#	snmpwalk -c <community> -OneU -v 2c <device>
# 
# Andrew Findlay
# andrew.findlay@skills-1st.co.uk
#
# 13 March 2003
########################################################################

# List of types as output by snmpwalk etc.
my $types = {
	'STRING' => ASN_OCTET_STR,
	'INTEGER' => ASN_INTEGER,
	'OID' => ASN_OBJECT_ID,
	'Timeticks' => ASN_TIMETICKS,
	'Gauge' => ASN_GAUGE,
	'Gauge32' => ASN_GAUGE,
	'Hex' => ASN_OCTET_STR,
	'Hex-STRING' => ASN_OCTET_STR,
	'Network Address' => ASN_OCTET_STR,
	'IpAddress' => ASN_IPADDRESS,
	'Counter32' => ASN_COUNTER,
	'Counter64' => ASN_COUNTER64,
};

########################################################################
# ProcessEntry()
# Stash away a 3-tuple (OID, TYPE, VALUE) - all are passed as text strings
########################################################################
sub ProcessEntry ($$$$) {
	my ($self, $oid, $type, $value) = @_;
	print "OID: $oid\nTYPE: $type\nVALUE: $value\n\n" if ($self->{debugging} > 2);

	# Derive data type from type-string
	my $datatype = $types->{$type};
	if (!defined($datatype)) {
		print STDERR "Unknown data type '$type' at line $.\n";
	}

	# Convert data value to correct form for re-use
	
	if ($type eq 'INTEGER') {
		# Value may have symbolic translation but we just want the number
		if ($value =~ /\(.*\)/) {
			($value) = ($value =~ /\(([0-9]+)\)/);
			# print "Converted INTEGER to $value\n";
		}
		# Value may have appended unit (e.g. '0 milliseconds') but we just want the number
		($value) = ($value =~ /([0-9]+)/);
	}

	if ($type eq 'Timeticks') {
		($value) = ($value =~ /\(([0-9]+)\)/);
		# print "Converted Timeticks to $value\n";
	}

	if (($type eq 'Hex-STRING') or ($type eq 'Hex')) {
		# Lowercase it all
		$value = lc $value;
		# Make sure each hex value has at least two digits
		$value =~ s/\s([0-9a-f])\s/ 0$1 /g;
		# Remove spaces
		$value =~ s/\s//g;
		# Convert hex string to string
		$value = pack 'H*', $value;
		# print "Converted Hex-STRING to '$value'\n";
	}

	if (($type eq 'STRING') and ($value =~ /^".*"$/)) {
		# KLUDGE! These are likely to be unreadable strings that would be
		# better shown as Hex-STRING
		# Just rip off the double quotes and drop through

		$value =~ s/^"//;
		$value =~ s/"$//;
	}

	if (($type eq 'STRING') and ($value =~ /^[0-9a-f:]+$/)) {
		# KLUDGE! These are really addresses that should be shown as Hex-STRING
		# Format: 0:4:c1:9c:90:c3

		my $newvalue = '';
		while ($value) {
			my $hex = $value;
			# Select just the first hex item
			$hex =~ s/:.*$//;
			# Remove first hex item from string
			$value =~ s/^[0-9a-f]+:*//i;
			$newvalue .= chr( hex($hex));
		}
		$value = $newvalue;
		# my $check = unpack 'H12', $value; print "Converted kludge-STRING to '$check'\n";
	}

	if ($type eq 'Network Address') {
		# These are of the form 0A:00:00:02
		$value =~ s/[\s:]//g;
		$value = pack 'H*', $value;
		# my $check = unpack 'H8', $value; print "Converted Network Address to $check\n";
	}

	if ($type eq 'IpAddress') {
		# These are of the form 10.0.0.253
		my @octets = split '\.', $value;
		$value = '';
		$value .= chr(shift(@octets));
		$value .= chr(shift(@octets));
		$value .= chr(shift(@octets));
		$value .= chr(shift(@octets));
		# my $check = unpack 'H8', $value; print "Converted IpAddress to $check\n";
	}

	# print "Storing OID $oid\n";

	# Record current OID using internal form
	push @{$self->{oidlist}}, new NetSNMP::OID($oid);

	# Note that we need to sort the list
	# In fact the input file may well be sorted anyway, but this feature is needed for
	# other purposes so we use it anyway
	$self->{oidlistDirty} = 1;

	# Get numeric form of OID
	my $oidstr = '.';
	$oidstr .= join '.', ${$self->{oidlist}}[-1]->to_array;

	# print "Storing OID $oid ($oidstr) type $datatype\n";

	# Create entry in hash
	$self->{mibdata}->{$oidstr} = {
		type => $datatype,
		value => $value,
	};
}

################################################################
#

=item C<< $agent->ParseDataFile($filename,$CompatKludges) >>

Read in SNMP OID = Value pairs from a file.  
Use this to load boiler-plate MIB items.

If $CompatKludges is true, tries to parse output from earlier
versions of Net-SNMP and UCD-SNMP. This is not completely successful,
as the output formats are ambiguous. It is strongly suggested
that $CompatKludges should be zero.

Returns zero if it fails to read the file, otherwise returns
the number of entries stored in the MIB.

=cut

#
################################################################

sub ParseDataFile ($$) {
	my ($self,$datafile,$CompatKludges) = @_;

	my ($oid, $type, $value);
	my ($oidLast, $typeLast, $valueLast);

	my $items = 0;

	print "Using data from $datafile\n" if ($self->{debugging});

	open SNMPDATA, $datafile or return 0;

	while (<SNMPDATA>) {
		chomp;

		# Ignore comments
		next if /^\s*#/;

		# Ignore blank lines
		next if /^\s*$/;

		################################################################
		# First try to fix some formatting problems in early versions of snmpwalk
		################################################################

		if ($CompatKludges) {
			# Watch for possible continuation lines and dont mess with them
			if (! /\.[^: =]+\s=/) {
				# Probably continuation
				# print "Probable continuation line: '$_'\n";
			}
			# Untyped text strings
			elsif (/\s=\s+".*"$/) {
				s/\s=\s+"/ = STRING: /;
				s/"$//;
				# print "Repaired string to: '$_'\n";
			}

			# Untyped integers (but we dont know what sort of int so make a guess)
			elsif (/\s=\s+[-]?[0-9]+$/) {
				s/\s=\s+([-]?[0-9]+)$/ = Counter32: $1/;
				# print "Repaired int to: '$_'\n";
			}

			# Untyped enumerated objects
			elsif (/\s=\s+.*\([0-9]+\)$/) {
				s/\s=\s+.*\(([0-9]+)\)$/ = INTEGER: $1/;
				# print "Repaired enum to: '$_'\n";
			}

			# Untyped MAC-address strings
			elsif (/\s=\s+[0-9a-f]+:[0-9a-f:]+$/i) {
				s/\s=\s+/ = STRING: /;
				# print "Repaired hex string to: '$_'\n";
			}

			# Remaining untyped lines could be anything so try them as strings
			elsif (! /\s=\s+[^: ]+:/) {
				s/\s=\s*/ = STRING: /;
				print "Guessed that untyped object should be: '$_'\n";
			}
		}

		################################################################
		# Now parse data from cleaned-up line
		################################################################

		# Most lines have a standard format
		# but continuation lines have to be guessed (they dont match this pattern)
		($oid, $type, $value) = /^([.0-9]+)\s+=\s+([^:]*):\s(.*)$/;

		# Allow for symbolic names
		if (!defined($oid)) {
			($oid, $type, $value) = /^(.+)\s+=\s+([^:]*):\s(.*)$/;
		}

		# No OID suggests continuation of previous data
		if (defined($oidLast) and !defined($oid)) {
			# Got a continuation line so add it to the previous value
			$valueLast .= "\n$_";
			next;
		}

		# Sanity check
		if (!defined($value)) {
			print "ERROR: undefined value in $_\n";
			$value = "";
		}

		if (defined($oid) and defined($oidLast)) {
			# Got a good line, so deal with the last one
			$self->ProcessEntry($oidLast, $typeLast, $valueLast);
			$items++;
			# Copy new data
			($oidLast, $typeLast, $valueLast) = ($oid, $type, $value);
			next;
		}

		if (defined($oid)) {
			($oidLast, $typeLast, $valueLast) = ($oid, $type, $value);
			next;
		}

		print STDERR "ODD ONE: $_\n";
	}

	# Handle last line
	if (defined($oidLast)) {
		$self->ProcessEntry($oidLast, $typeLast, $valueLast);
		$items++;
	}

	close SNMPDATA;

	####### # test
	####### my $item;
	####### 
	####### $item = $mibdata{'.1.3.6.1.2.1.1.1.0'};
	####### printf "OID: %s\nPREV: %s\nNEXT: %s\nTYPE: %s\nVALUE: %s\n\n", $item->{'OID'},
	####### $item->{'PrevOID'}, $item->{'NextOID'}, $item->{'type'}, $item->{'value'};
	####### $item = $mibdata{'.1.3.6.1.2.1.1.2.0'};
	####### printf "OID: %s\nPREV: %s\nNEXT: %s\nTYPE: %s\nVALUE: %s\n\n", $item->{'OID'},
	####### $item->{'PrevOID'}, $item->{'NextOID'}, $item->{'type'}, $item->{'value'};
	####### $item = $mibdata{'.1.3.6.1.2.1.47.1.4.1.0'};
	####### printf "OID: %s\nPREV: %s\nNEXT: %s\nTYPE: %s\nVALUE: %s\n\n", $item->{'OID'},
	####### $item->{'PrevOID'}, $item->{'NextOID'}, $item->{'type'}, $item->{'value'};

	# if ($self->{debugging}) {
	# 	my $n;
	# 	for ($n=0; defined($oidlist[$n]); $n++) {
	# 		print "OID $n: $oidlist[$n] -- ";
	# 		my $oidstr .= join '.', $oidlist[$n]->to_array;
	# 		print "($oidstr)\n";
	# 	}
	# }

	# Might as well do this now, though it may well get done again later
	$self->BuildOIDList();

	return $items;
}

########################################################################
# End of parsing code
########################################################################

########################################################################
# BuildOIDList()
#
# Rebuilids the OID list with all keys in ascending order if this
# has not been done since the last build.
########################################################################

sub BuildOIDList ($) {
	my $self = shift;

	return if (! $self->{oidlistDirty});

	print "About to build the OID list\n" if ($self->{debugging});

	# NetSNMP::OID overloads the <=> operator so this should be easy
	# but it just does not work if left without a specific comparison function
	@{$self->{oidlist}} = sort MyComp @{$self->{oidlist}};

	# Now re-build the index
	undef %{$self->{oidlistindex}};

	my $n;
	for ($n=0; defined(${$self->{oidlist}}[$n]); $n++) {
		my $thisoid = '.';
		$thisoid .= join '.', ${$self->{oidlist}}[$n]->to_array;
		${$self->{oidlistindex}}{$thisoid} = $n;
		# print "Indexed $thisoid at $n\n";
	}

	$self->{oidlistDirty} = 0;
	print "Sorted the OID list\n" if ($self->{debugging});
}

# Comparison routine for sort
#
sub MyComp($$) {
	my ($l, $r) = @_;

	# print "In MyComp\n";

	my $res = ($l <=> $r);

	# print "Comparing $l with $r: $res\n";
	return $res;
}

########################################################################
#

=item C<< $agent->version >>

Returns the version number of the Net::Raddle::SNMPAgent module

=cut

#
########################################################################

sub version ($) {
	return $VERSION;
}


########################################################################
########################################################################
# Procedures for customising the emulated device

=back

=head2 Utility Procedures

B<NOTE> these are not exported by default.
You should either import them explicitly with:

	use Net::Raddle::SNMPAgent qw(IPToMAC Netmask CanonicalOID);

or you should refer to them using the package name, e.g.:

	my $mask = Net::Raddle::SNMPAgent::Netmask(27);

=over

=cut

#
########################################################################
########################################################################

########################################################################
# AssertIP($IPAddress)
#
# Throws an exception if $IPAddress is not a valid IP address
# We are quite strict here: there must be four octets, each with
# a value between 0 and 255 inclusive, and there must be no leading zeros.
#
########################################################################

sub AssertIP ($) {
	my ($IP) = @_;

	my @octet = ($IP =~ /^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$/);

	if (((scalar @octet) != 4) ||
			($octet[0] > 255) ||
			($octet[1] > 255) ||
			($octet[2] > 255) ||
			($octet[3] > 255)) {
		confess "Bad IP address '$IP'";
	}

	if (		($octet[0] =~ /^0./) ||
			($octet[1] =~ /^0./) ||
			($octet[2] =~ /^0./) ||
			($octet[3] =~ /^0./)) {
		confess "Leading zeros not permitted in IP address '$IP'";
	}
}


########################################################################
#

=item CanonicalOID($oid)

Returns canonical text string form of an OID. e.g.:

	my $oidstr = CanonicalOID( 'IF-MIB::ifNumber.0' );

Canonical OIDs have the form: '.1.3.6.1.2.1.1.3.0'

Returns I<undef> if the OID cannot be processed.

=cut

#
########################################################################

sub CanonicalOID ($) {
	my ($in) = @_;

	my $oid = new NetSNMP::OID($in);
	return undef if (!$oid);

	return '.' . join '.', $oid->to_array;
}

######################################################################
#

=item IPToMAC($prefix,$IP)

Invent a likely MAC address from an IP address and a 2-octet prefix.
e.g.:

	my $MACaddr = IPToMac( '00:30', '10.3.76.45' );

This allows the ARP tables of different devices to be set up with
consistent data without having to hand-configure the mappings.

=cut

#
######################################################################

sub IPToMAC ($$) {
	my ($Prefix, $IP) = @_;

	# print "IPToMAC( $Prefix, $IP ) -> ";

	# Check parameters
	if ($Prefix !~ /^[0-9a-fA-F][0-9a-fA-F]:[0-9a-fA-F][0-9a-fA-F]$/) {
		confess "Bad prefix format '$Prefix'";
	}
	AssertIP($IP);

	my $MAC = "$Prefix:";

	my $hash .= md5_hex($IP);
	$hash =~ s/(..)(..)(..)(..).*/$1:$2:$3:$4/;

	$MAC .= $hash;

	# print "$MAC\n";

	return $MAC;
}

######################################################################
#

=item Netmask($length)

Convert netmask length to dotted-quad form, e.g.:

	my $mask = Netmask(24);
	# $mask is now '255.255.255.0'

=cut

#
######################################################################
sub Netmask ($) {
	my ($slashnum) = @_;

	confess "Bad value passed to Netmask" if (!$slashnum or $slashnum < 4 or $slashnum > 32);

	my $bitmask = "";
	my $n;
	for ($n=0; $n < 32; $n++) {
		if ($n < $slashnum) {
			$bitmask .= "1";
		}
		else {
			$bitmask .= "0";
		}
	}

	my $binmask = pack "B32", $bitmask;
	my ($a1, $a2, $a3, $a4) = unpack "aaaa", $binmask;
	$a1 = ord $a1;
	$a2 = ord $a2;
	$a3 = ord $a3;
	$a4 = ord $a4;

	return "$a1.$a2.$a3.$a4";
}



########################################################################
#
=item IPTo4Bytes( IPaddr )

 $Addr = IPTo4Bytes( '10.0.0.177' );

 Converts text form of IP address to binary string

=cut
#
########################################################################

sub IPTo4Bytes ($) {
	my ($value) = @_;

	AssertIP($value);

	my @octets = split '\.', $value;
	$value = '';
	$value .= chr(shift(@octets));
	$value .= chr(shift(@octets));
	$value .= chr(shift(@octets));
	$value .= chr(shift(@octets));

	return $value;
}

########################################################################
#
=item DeHex

Convert a hex string into a binary string

	$binStr = DeHex( '0:30:2a:45:9b:aa' );
	$binStr = DeHex( 'A0 30 4 0a 56 cc' );

=cut
#
########################################################################
sub DeHex ($) {
	my ($value) = @_;

	# Lowercase it all and add leading and trailing markers
	$value = ':' . (lc $value) . ':';
	# Make sure each hex value has at least two digits
	$value =~ s/[:\s]([0-9a-f])[:\s]/ 0$1 /g;
	# Remove spaces and colons
	$value =~ s/[:\s]//g;
	# Convert hex string to string
	$value = pack 'H*', $value;
	# my $hs = unpack 'H*', $value;
	# print "DeHexed to '$hs'\n";

	return $value;
}

########################################################################
#
=item ToHex

Convert a binary string to a colon-separated hex string

	$HexString = ToHex( $binstring );

=cut
#
########################################################################

sub ToHex ($) {
	my ($value) = @_;

	my $res = unpack 'H*', $value;
	$res =~ s/(..)/$1:/g;
	$res =~ s/:$//g;
	return $res;
}

########################################################################
#
=item HexToOIDpart

Convert ia hex string to a string of decimals for use as part of an OID.

	$Oidpart = IPtoOIDpart( '00:20:AF:5B:82:CE' );
	$Oidpart = IPtoOIDpart( '0 20 af' );

Result is of the form:

	'0.32.175'
=cut
#
########################################################################

sub HexToOIDpart ($) {
	my ($value) = @_;

	# Convert the hex string to a byte string then unpack it into a list of integers
	my @bytes = unpack 'C*', DeHex( $value );

	# Add each byte value to the result string
	my $res = '';
	while (defined($bytes[0])) {
		$res .= sprintf "%d.", shift @bytes;
	}

	# Remove the extra '.' from the end
	$res =~ s/\.$//;

	return $res;
}

# print "HexToOIDpart('07:02:FF'): ", HexToOIDpart('07:02:FF'), "\n";


########################################################################
########################################################################
#

=back

=head2 Procedures acting on individual MIB items

=over

=cut

#
########################################################################
########################################################################

########################################################################
#

=item C<< $agent->TieMIBItem($canonicaloid,$type,$classname,$classparam) >>

Ties a MIB item to a class, creating the item if necessary.
The MIB item to be tied is specified as a canonical OID and an
SNMP data type. e.g.:

	$agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'), ASN_TIMETICKS,
		"TicksSince", Mktime( 2002, 9, 24, 3, 59, 14) );

Available data types are:

	ASN_OBJECT_ID	An OID - must be in canonical form.
	ASN_OCTET_STR	Octet string. Can be text or binary.
	ASN_INTEGER	32-bit integer.
	ASN_TIMETICKS	Time ticks (units of 0.01s)
	ASN_GAUGE	32-bit non-wrapping integer.
	ASN_IPADDRESS	An IP address in the form 10.3.27.200
	ASN_COUNTER	32-bit counter.
	ASN_COUNTER64	64-bit counter (not yet working properly).

=cut

#
########################################################################

sub TieMIBItem ($$$$$) {
	my ($self, $canonicalOID, $type, $classname, $classparam) = @_;

	# Sanity check
	if (not $canonicalOID) {
		carp "OID not defined for TieMIBItem";
	}
	if ($canonicalOID !~ /^\.[0-9]+[0-9.]*$/) {
		carp "TieMIBItem needs a canonical OID: '$canonicalOID'";
	}

	if (!defined( $self->{mibdata}->{$canonicalOID} )) {
		# Record OID using internal form
		push @{$self->{oidlist}}, new NetSNMP::OID($canonicalOID);

		# Note that we need to sort the list
		$self->{oidlistDirty} = 1;

		# Create entry in hash
		$self->{mibdata}->{$canonicalOID} = {
			type => $type,
		};
	}

	# Now tie the class into the item
	tie $self->{mibdata}->{$canonicalOID}->{'value'}, $classname, $classparam
		or confess "TieMIBItem($canonicalOID, $type, $classname, $classparam) failed";
}

########################################################################
#

=item C<< $agent->SetMIBValue($canonicaloid,$type,$initialvalue) >>

Sets the value of a MIB item, creating the item if it does not
already exist. e.g.:

	SetMIBValue( '.1.3.6.1.2.1.1.5.0', ASN_OCTET_STR, 'router.example.org' )

Throws an exception if the OID is missing or unusable.
=cut

#
########################################################################

sub SetMIBValue ($$$$) {
	my ($self, $canonicalOID, $type, $value) = @_;

	# Sanity check
	if (not $canonicalOID) {
		croak "OID not defined for SetMIBValue";
	}
	if ($canonicalOID !~ /^\.[0-9]+[0-9.]*$/) {
		croak "Not a canonical OID: '$canonicalOID'";
	}

	if (defined( $self->{mibdata}->{$canonicalOID} )) {
		$self->{mibdata}->{$canonicalOID}->{'value'} = $value;
	}
	else {
		# Record OID using internal form
		push @{$self->{oidlist}}, new NetSNMP::OID($canonicalOID);

		# Note that we need to sort the list
		$self->{oidlistDirty} = 1;

		# Create entry in hash
		$self->{mibdata}->{$canonicalOID} = {
			type => $type,
			value => $value,
		};
	}
}


########################################################################
#

=item C<< $agent->GetMIBValue($canonicaloid) >>

Gets the value of a MIB item.

Returns a hash of the form:

	{ type => DATATYPE, value => VALUE }

Returns I<undef> if the item is not defined.
Note that VALUE may be a closure or function, in which case it may need to be called
to obtain a usable value.

	my %mibitem = $agent->GetMIBValue( '.1.3.6.1.2.1.1.5.0' );

=cut

#
########################################################################

sub GetMIBValue ($$) {
	my ($self, $canonicalOID) = @_;

	# Sanity check
	if ($canonicalOID !~ /^\.[0-9]+[0-9.]*$/) {
		confess "Not a canonical OID: '$canonicalOID'";
	}

	return $self->{mibdata}->{$canonicalOID};
}

########################################################################
########################################################################
#

=back

=head2 Procedures working on higher-level objects

=over

=cut

#
########################################################################
########################################################################

########################################################################
#

=item C<< $agent->AddIF($datahash) >>

Add an interface, with data as specified in C<$datahash> e.g.:

	$IFIndex = $agent->AddIF( { ifType => 'csmacd', ifSpeed => 10000000 } );

The complete list of values that can be specified is the ifTable:

        ifIndex                 InterfaceIndex,
        ifDescr                 DisplayString,
        ifType                  IANAifType,
        ifMtu                   Integer32,
        ifSpeed                 Gauge32,
        ifPhysAddress           PhysAddress,
        ifAdminStatus           INTEGER,
        ifOperStatus            INTEGER,
        ifLastChange            TimeTicks,
        ifInOctets              Counter32,
        ifInUcastPkts           Counter32,
        ifInNUcastPkts          Counter32,  -- deprecated
        ifInDiscards            Counter32,
        ifInErrors              Counter32,
        ifInUnknownProtos       Counter32,
        ifOutOctets             Counter32,
        ifOutUcastPkts          Counter32,
        ifOutNUcastPkts         Counter32,  -- deprecated
        ifOutDiscards           Counter32,
        ifOutErrors             Counter32,
        ifOutQLen               Gauge32,    -- deprecated
        ifSpecific              OBJECT IDENTIFIER -- deprecated

These items from ifXTable may also be specified:

        ifName                  DisplayString,
        ifInMulticastPkts       Counter32,
        ifInBroadcastPkts       Counter32,
        ifOutMulticastPkts      Counter32,
        ifOutBroadcastPkts      Counter32,
        ifAlias                 DisplayString,
	ifLinkUpDownTrapEnable	Integer,
	ifHighSpeed		Gauge32,
	ifPromiscuousMode	Integer,
	ifConnectorPresent	Integer

Also these from the ifStackTable:

	ifStackStatus		-- The two minimal entries will be created
				-- unless the value 'no' is explicitly set.

Any value that is not supplied will be defaulted.

Include 'WithDeprecatedItems => 1' in C<$datahash> if you want
to return data for the items marked as deprecated.

=cut

#
########################################################################

sub AddIF ($$) {
	my ($self, $param) = @_;

	print "AddIF type $param->{'ifType'}\n" if ($self->{debugging});

	my $WithDeprecatedItems = $param->{'WithDeprecatedItems'};

	# Bump the number of interfaces and remember the result
	my $ifNumberOID = CanonicalOID('IF-MIB::ifNumber.0');
	if (!defined($self->{mibdata}->{$ifNumberOID})) {
		# Need to create the MIB item from scratch
		$self->SetMIBValue( $ifNumberOID, ASN_INTEGER, 0);
	}
	my $ifn = ++$self->{mibdata}->{$ifNumberOID}->{'value'};

	print "ifNumber.0 now $ifn\n" if ($self->{debugging});

	# Index entry
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifIndex.$ifn"), ASN_INTEGER, $ifn );

	# ifDescr
	my $descr = $param->{'ifDescr'};
	$descr = "" if (!defined($descr));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifDescr.$ifn"), ASN_OCTET_STR, $descr );

	# ifType
	my $ifType = $param->{'ifType'};
	$ifType = 6 if (!defined($ifType));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifType.$ifn"), ASN_INTEGER, $ifType );

	# ifMtu
	my $ifMtu = $param->{'ifMtu'};
	$ifMtu = 1500 if (!defined($ifMtu));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifMtu.$ifn"), ASN_INTEGER, $ifMtu );

	# ifSpeed
	my $ifSpeed = $param->{'ifSpeed'};
	$ifSpeed = 10000000 if (!defined($ifSpeed));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifSpeed.$ifn"), ASN_GAUGE, $ifSpeed );

	# ifPhysAddress
	my $ifPhysAddress = $param->{'ifPhysAddress'};
	$ifPhysAddress = "" if (!defined($ifPhysAddress));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifPhysAddress.$ifn"), ASN_OCTET_STR, DeHex($ifPhysAddress) );

	# ifAdminStatus
	my $ifAdminStatus = $param->{'ifAdminStatus'};
	$ifAdminStatus = 1 if (!defined($ifAdminStatus));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifAdminStatus.$ifn"), ASN_INTEGER, $ifAdminStatus );

	# ifOperStatus
	my $ifOperStatus = $param->{'ifOperStatus'};
	$ifOperStatus = 1 if (!defined($ifOperStatus));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOperStatus.$ifn"), ASN_INTEGER, $ifOperStatus );

	# ifLastChange
	my $ifLastChange = $param->{'ifLastChange'};
	$ifLastChange = 0 if (!defined($ifLastChange));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifLastChange.$ifn"), ASN_TIMETICKS, $ifLastChange );

	# ifInOctets
	my $ifInOctets = $param->{'ifInOctets'};
	$ifInOctets = 0 if (!defined($ifInOctets));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInOctets.$ifn"), ASN_COUNTER, $ifInOctets );

	# ifInUcastPkts
	my $ifInUcastPkts = $param->{'ifInUcastPkts'};
	$ifInUcastPkts = 0 if (!defined($ifInUcastPkts));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInUcastPkts.$ifn"), ASN_COUNTER, $ifInUcastPkts );

	# ifInNUcastPkts is deprecated
	if ($WithDeprecatedItems) {
		my $ifInNUcastPkts = $param->{'ifInNUcastPkts'};
		$ifInNUcastPkts = 0 if (!defined($ifInNUcastPkts));
		$self->SetMIBValue( CanonicalOID("IF-MIB::ifInNUcastPkts.$ifn"), ASN_COUNTER, $ifInNUcastPkts );
	}

	# ifInDiscards
	my $ifInDiscards = $param->{'ifInDiscards'};
	$ifInDiscards = 0 if (!defined($ifInDiscards));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInDiscards.$ifn"), ASN_COUNTER, $ifInDiscards );

	# ifInErrors
	my $ifInErrors = $param->{'ifInErrors'};
	$ifInErrors = 0 if (!defined($ifInErrors));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInErrors.$ifn"), ASN_COUNTER, $ifInErrors );

	# ifInUnknownProtos
	my $ifInUnknownProtos = $param->{'ifInUnknownProtos'};
	$ifInUnknownProtos = 0 if (!defined($ifInUnknownProtos));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInUnknownProtos.$ifn"), ASN_COUNTER, $ifInUnknownProtos );

	# ifOutOctets
	my $ifOutOctets = $param->{'ifOutOctets'};
	$ifOutOctets = 0 if (!defined($ifOutOctets));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutOctets.$ifn"), ASN_COUNTER, $ifOutOctets );

	# ifOutUcastPkts
	my $ifOutUcastPkts = $param->{'ifOutUcastPkts'};
	$ifOutUcastPkts = 0 if (!defined($ifOutUcastPkts));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutUcastPkts.$ifn"), ASN_COUNTER, $ifOutUcastPkts );

	# ifOutNUcastPkts
	if ($WithDeprecatedItems) {
		my $ifOutNUcastPkts = $param->{'ifOutNUcastPkts'};
		$ifOutNUcastPkts = 0 if (!defined($ifOutNUcastPkts));
		$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutNUcastPkts.$ifn"), ASN_COUNTER, $ifOutNUcastPkts );
	}

	# ifOutDiscards
	my $ifOutDiscards = $param->{'ifOutDiscards'};
	$ifOutDiscards = 0 if (!defined($ifOutDiscards));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutDiscards.$ifn"), ASN_COUNTER, $ifOutDiscards );

	# ifOutErrors
	my $ifOutErrors = $param->{'ifOutErrors'};
	$ifOutErrors = 0 if (!defined($ifOutErrors));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutErrors.$ifn"), ASN_COUNTER, $ifOutErrors );

	# ifOutQLen
	if ($WithDeprecatedItems) {
		my $ifOutQLen = $param->{'ifOutQLen'};
		$ifOutQLen = 0 if (!defined($ifOutQLen));
		$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutQLen.$ifn"), ASN_GAUGE, $ifOutQLen );
	}

	# ifSpecific
	if ($WithDeprecatedItems) {
		my $ifSpecific = $param->{'ifSpecific'};
		$ifSpecific = '.0.0' if (!defined($ifSpecific));
		$self->SetMIBValue( CanonicalOID("IF-MIB::ifSpecific.$ifn"), ASN_OBJECT_ID, $ifSpecific );
	}

	####################### Now ifXTable items ################################################

	# ifName
	my $ifName = $param->{'ifName'};
	$ifName = "" if (!defined($ifName));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifName.$ifn"), ASN_OCTET_STR, $ifName );

	# ifInMulticastPkts
	my $ifInMulticastPkts = $param->{'ifInMulticastPkts'};
	$ifInMulticastPkts = 0 if (!defined($ifInMulticastPkts));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInMulticastPkts.$ifn"), ASN_COUNTER, $ifInMulticastPkts );

	# ifInBroadcastPkts
	my $ifInBroadcastPkts = $param->{'ifInBroadcastPkts'};
	$ifInBroadcastPkts = 0 if (!defined($ifInBroadcastPkts));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifInBroadcastPkts.$ifn"), ASN_COUNTER, $ifInBroadcastPkts );

	# ifOutMulticastPkts
	my $ifOutMulticastPkts = $param->{'ifOutMulticastPkts'};
	$ifOutMulticastPkts = 0 if (!defined($ifOutMulticastPkts));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutMulticastPkts.$ifn"), ASN_COUNTER, $ifOutMulticastPkts );

	# ifOutBroadcastPkts
	my $ifOutBroadcastPkts = $param->{'ifOutBroadcastPkts'};
	$ifOutBroadcastPkts = 0 if (!defined($ifOutBroadcastPkts));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifOutBroadcastPkts.$ifn"), ASN_COUNTER, $ifOutBroadcastPkts );

	# ifAlias
	my $ifAlias = $param->{'ifAlias'};
	$ifAlias = "" if (!defined($ifAlias));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifAlias.$ifn"), ASN_OCTET_STR, $ifAlias );

	# ifConnectorPresent
	# default true(1)
	my $ifConnectorPresent = $param->{'ifConnectorPresent'};
	$ifConnectorPresent = 1 if (!defined($ifConnectorPresent));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifConnectorPresent.$ifn"), ASN_INTEGER, $ifConnectorPresent );

	# ifPromiscuousMode
	# default false(2)
	my $ifPromiscuousMode = $param->{'ifPromiscuousMode'};
	$ifPromiscuousMode = 2 if (!defined($ifPromiscuousMode));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifPromiscuousMode.$ifn"), ASN_INTEGER, $ifPromiscuousMode );

	# ifLinkUpDownTrapEnable
	# default true(1)
	my $ifLinkUpDownTrapEnable = $param->{'ifLinkUpDownTrapEnable'};
	$ifLinkUpDownTrapEnable = 1 if (!defined($ifLinkUpDownTrapEnable));
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifLinkUpDownTrapEnable.$ifn"), ASN_INTEGER, $ifLinkUpDownTrapEnable );

	# ifHighSpeed
	# default ifSpeed/10**6 if ifSpeed is set
	my $ifHighSpeed = $param->{'ifHighSpeed'};
	if (!defined($ifHighSpeed)) {
		$ifHighSpeed = 0;
		if ($ifSpeed) {
			$ifHighSpeed = int( $ifSpeed / 1000000 );
		}
	}
	$self->SetMIBValue( CanonicalOID("IF-MIB::ifHighSpeed.$ifn"), ASN_GAUGE, $ifHighSpeed );

	####################### Now ifStackTable items ################################################

	# ifStackStatus
	# Here we just create two standard entries saying that nothing is stacked
	# above or below this interface (unless told not to do so)
	#
	if (!defined($param->{'ifStackStatus'}) or ($param->{'ifStackStatus'} ne 'no')) {
		$self->SetMIBValue( CanonicalOID("IF-MIB::ifStackStatus.0.$ifn"), ASN_INTEGER, 1);
		$self->SetMIBValue( CanonicalOID("IF-MIB::ifStackStatus.$ifn.0"), ASN_INTEGER, 1);
	}


	# Return ifIndex
	return $ifn;
}

########################################################################
#

=item C<< $agent->AddIpAddrEntry($datahash) >>

Add an IpAddrEntry - this is what binds IP addresses to interfaces

e.g.:

	$agent->AddIpAddrEntry( {
		ipAdEntAddr => '10.0.0.177',
		ipAdEntIfIndex => 2,
		ipAdEntNetMask => '255.255.255.0',
		ipAdEntBcastAddr => 1,
		ipAdEntReasmMaxSize => 8192,
	} );

ipAdEntAddr and ipAdEntIfIndex MUST be specified.
Other values default as shown above if not specified.

=cut

#
########################################################################

sub AddIpAddrEntry ($$) {
	my ($self, $param) = @_;

	# Sanity check
	if (!$param->{'ipAdEntAddr'}) {
		confess "AddIpAddrEntry called without an ipAdEntAddr";
	}
	if (!$param->{'ipAdEntIfIndex'}) {
		confess "AddIpAddrEntry called without an ipAdEntIfIndex";
	}

	print "AddIpAddrEntry  $param->{'ipAdEntAddr'}\n" if ($self->{debugging});

        # ipAdEntAddr
        my $ipAdEntAddr = $param->{'ipAdEntAddr'};
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipAdEntAddr.$ipAdEntAddr"), ASN_IPADDRESS, IPTo4Bytes($ipAdEntAddr) );

        # ipAdEntIfIndex
        my $ipAdEntIfIndex = $param->{'ipAdEntIfIndex'};
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipAdEntIfIndex.$ipAdEntAddr"), ASN_INTEGER, $ipAdEntIfIndex );

        # ipAdEntNetMask
        my $ipAdEntNetMask = $param->{'ipAdEntNetMask'};
        $ipAdEntNetMask = '255.255.255.0' if (!defined($ipAdEntNetMask));
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipAdEntNetMask.$ipAdEntAddr"), ASN_IPADDRESS, IPTo4Bytes($ipAdEntNetMask) );

        # ipAdEntBcastAddr
        my $ipAdEntBcastAddr = $param->{'ipAdEntBcastAddr'};
        $ipAdEntBcastAddr = 1 if (!defined($ipAdEntBcastAddr));
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipAdEntBcastAddr.$ipAdEntAddr"), ASN_INTEGER, $ipAdEntBcastAddr );

        # ipAdEntReasmMaxSize
        my $ipAdEntReasmMaxSize = $param->{'ipAdEntReasmMaxSize'};
        $ipAdEntReasmMaxSize = 8192 if (!defined($ipAdEntReasmMaxSize));
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipAdEntReasmMaxSize.$ipAdEntAddr"), ASN_INTEGER, $ipAdEntReasmMaxSize );
}

########################################################################
#

=item C<< $agent->AddIpNetToMediaEntry($datahash) >>

Add an IpNetToMediaEntry - this is what binds IP addresses to MAC addresses
(the ARP table). e.g.:

	$agent->AddIpNetToMediaEntry( {
       		ipNetToMediaIfIndex => 3,
       		ipNetToMediaPhysAddress => '0:30:b6:36:14:d4',
       		ipNetToMediaNetAddress => '10.0.0.177',
       		ipNetToMediaType => 4,
	} );

The first three MUST be specified.
ipNetToMediaType defaults to type dynamic(3) if not specified
Possible values are:

       other(1),        -- none of the following
       invalid(2),      -- an invalidated mapping
       dynamic(3),	-- obtained by dynamic means (e.g. ARP)
       static(4)	-- static (configuration or hardware)

=cut

#
########################################################################

sub AddIpNetToMediaEntry ($$) {
	my ($self, $param) = @_;

	# Sanity check
	if (!$param->{'ipNetToMediaIfIndex'}) {
		confess "AddIpAddrEntry called without an ipNetToMediaIfIndex";
	}
	if (!$param->{'ipNetToMediaPhysAddress'}) {
		confess "AddIpAddrEntry called without an ipNetToMediaPhysAddress";
	}
	if (!$param->{'ipNetToMediaNetAddress'}) {
		confess "AddIpAddrEntry called without an ipNetToMediaNetAddress";
	}

	print "AddIpNetToMediaEntry  $param->{'ipNetToMediaNetAddress'} $param->{'ipNetToMediaIfIndex'} $param->{'ipNetToMediaPhysAddress'}\n" if ($self->{debugging});

        # ipNetToMediaNetAddress
        my $ipNetToMediaNetAddress = $param->{'ipNetToMediaNetAddress'};
	AssertIP($ipNetToMediaNetAddress);
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipNetToMediaNetAddress.$ipNetToMediaNetAddress"),
		ASN_IPADDRESS, IPTo4Bytes($ipNetToMediaNetAddress) );

        # ipNetToMediaIfIndex
        my $ipNetToMediaIfIndex = $param->{'ipNetToMediaIfIndex'};
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipNetToMediaIfIndex.$ipNetToMediaNetAddress"),
		ASN_INTEGER, $ipNetToMediaIfIndex );

        # ipNetToMediaType
        my $ipNetToMediaType = $param->{'ipNetToMediaType'};
	$ipNetToMediaType = 3 if (!defined($ipNetToMediaType));
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipNetToMediaType.$ipNetToMediaNetAddress"),
		ASN_INTEGER, $ipNetToMediaType );

        # ipNetToMediaPhysAddress
        my $ipNetToMediaPhysAddress = $param->{'ipNetToMediaPhysAddress'};
        $self->SetMIBValue( CanonicalOID("IP-MIB::ipNetToMediaPhysAddress.$ipNetToMediaNetAddress"),
		ASN_OCTET_STR, DeHex($ipNetToMediaPhysAddress) );
}

########################################################################
#

=item C<< $agent->AddBridgePort($datahash) >>

Add a dot1d bridge port entry to an existing interface

	$portID = $agent->AddBridgePort( {
		dot1dBasePort => 42,
		dot1dBasePortIfIndex => 5,
		dot1dBasePortCircuit => oidzero,
		dot1dBasePortDelayExceededDiscards => 0,
		dot1dBasePortMtuExceededDiscards => 0,
	} );

dot1dBasePortIfIndex MUST be specified.

=cut

#
########################################################################

sub AddBridgePort ($$) {
	my ($self, $param) = @_;

	# Sanity check
	if (!$param->{'dot1dBasePortIfIndex'}) {
		confess "AddBridgePort called without an dot1dBasePortIfIndex";
	}

	# Bump the number of ports and remember the result
	my $portNumberOID = CanonicalOID('BRIDGE-MIB::dot1dBaseNumPorts.0');
	if (!defined($self->{mibdata}->{$portNumberOID})) {
		# Need to create the MIB item from scratch
		$self->SetMIBValue( $portNumberOID, ASN_INTEGER, 0);
	}
	my $ptn = ++$self->{mibdata}->{$portNumberOID}->{'value'};

	print "portNumber.0 now $ptn\n" if ($self->{debugging});

	##################### build the Dot1dBasePortEntry ###################

	# Index entry
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBasePort.$ptn"), ASN_INTEGER, $ptn );

	# The interface that this port is associated with
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBasePortIfIndex.$ptn"), ASN_INTEGER, $param->{'dot1dBasePortIfIndex'});

	# The circuit that this port is associated with (usually OID 0.0 for simple interfaces)
	my $circuit = '.0.0';
	$circuit = $param->{'dot1dBasePortCircuit'} if ($param->{'dot1dBasePortCircuit'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBasePortCircuit.$ptn"), ASN_OBJECT_ID, $circuit);

	# dot1dBasePortDelayExceededDiscards
	my $dot1dBasePortDelayExceededDiscards = 0;
	$dot1dBasePortDelayExceededDiscards = $param->{'dot1dBasePortDelayExceededDiscards'}
		if ($param->{'dot1dBasePortDelayExceededDiscards'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBasePortDelayExceededDiscards.$ptn"),
		ASN_COUNTER, $dot1dBasePortDelayExceededDiscards);

	# dot1dBasePortMtuExceededDiscards
	my $dot1dBasePortMtuExceededDiscards = 0;
	$dot1dBasePortMtuExceededDiscards = $param->{'dot1dBasePortMtuExceededDiscards'}
		if ($param->{'dot1dBasePortMtuExceededDiscards'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBasePortMtuExceededDiscards.$ptn"),
		ASN_COUNTER, $dot1dBasePortMtuExceededDiscards);

	# Return the bridge port ID
	return $ptn;
}


########################################################################
#

=item C<< $agent->SetSpanningTree($datahash) >>

Sets the basic Spanning Tree parameters for a bridge instance

	$agent->SetSpanningTree( {
		dot1dStpDesignatedRoot => '00:0C:4A:27:65:BF',
	} );

=cut

#
########################################################################

sub SetSpanningTree ($$) {
	my ($self, $param) = @_;

	# dot1dStpDesignatedRoot
	if ($param->{'dot1dStpDesignatedRoot'}) {
		my $dot1dStpDesignatedRoot = $param->{'dot1dStpDesignatedRoot'};
		$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpDesignatedRoot.0"),
			ASN_OCTET_STR, DeHex($dot1dStpDesignatedRoot) );
	}
}


########################################################################
#

=item C<< $agent->AddStpPortEntry($datahash) >>

Add a spanning tree port entry to an existing bridge port e.g.:

	$portID = $agent->AddStpPortEntry( {
		  dot1dStpPort => 5,
                  dot1dStpPortPriority => 128,
                  dot1dStpPortState => 5,
                  dot1dStpPortEnable => 1,
                  dot1dStpPortPathCost => 19,
                  dot1dStpPortDesignatedRoot => '00:2A:00:C0:92:73:A7:C0',
                  dot1dStpPortDesignatedCost => 73,
                  dot1dStpPortDesignatedBridge => '00:2A:00:C0:92:73:A7:C0',
                  dot1dStpPortDesignatedPort => '80:15',
                  dot1dStpPortForwardTransitions => 86,
	} );

dot1dStpPort MUST be specified.
If other params are left undefined they will be defaulted by assuming
that this is the only (or root) bridge in the domain.

NOTE: SetBridge must be called first, to set dot1dBaseBridgeAddress
before calling AddStpPortEntry.

=cut

#
########################################################################

sub AddStpPortEntry ($$) {
	my ($self, $param) = @_;

	# Sanity check
	if (!$param->{'dot1dStpPort'}) {
		confess "AddStpPortEntry called without a dot1dStpPort";
	}
	my $port = $param->{'dot1dStpPort'};

	print "AddStpPortEntry $port\n" if ($self->{debugging});

	##################### build the Dot1dStpPortEntry ###################

	# Find and stash the bridge address
	my $BridgeBaseOID = CanonicalOID("BRIDGE-MIB::dot1dBaseBridgeAddress.0");
        if (!defined($self->{mibdata}->{$BridgeBaseOID})) {
		confess "AddStpPortEntry called but dot1dBaseBridgeAddress has not been set yet";
	}
	my $BridgeBaseMAC = ToHex($self->{mibdata}->{$BridgeBaseOID}->{'value'});
	print "BridgeBaseMAC: $BridgeBaseMAC\n" if ($self->{debugging});


	# Index entry
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBasePortIfIndex.$port"), ASN_INTEGER, $port );

	# dot1dStpPortPriority
	my $dot1dStpPortPriority = 128;
	$dot1dStpPortPriority = $param->{'dot1dStpPortPriority'}
		if ($param->{'dot1dStpPortPriority'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortPriority.$port"),
		ASN_INTEGER, $dot1dStpPortPriority);

	# dot1dStpPortState
	# default to forwarding(5)
	my $dot1dStpPortState = 5;
	$dot1dStpPortState = $param->{'dot1dStpPortState'}
		if ($param->{'dot1dStpPortState'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortState.$port"),
		ASN_INTEGER, $dot1dStpPortState);

	# dot1dStpPortEnable
	# default to enable(1)
	my $dot1dStpPortEnable = 1;
	$dot1dStpPortEnable = $param->{'dot1dStpPortEnable'}
		if ($param->{'dot1dStpPortEnable'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortEnable.$port"),
		ASN_INTEGER, $dot1dStpPortEnable);

	# dot1dStpPortPathCost
	my $dot1dStpPortPathCost = 100;
	$dot1dStpPortPathCost = $param->{'dot1dStpPortPathCost'}
		if ($param->{'dot1dStpPortPathCost'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortPathCost.$port"),
		ASN_INTEGER, $dot1dStpPortPathCost);

	# dot1dStpPortDesignatedRoot
	# Default is '00:2A' followed by the MAC address of this bridge
	my $dot1dStpPortDesignatedRoot = '00:2A:' . $BridgeBaseMAC;
	$dot1dStpPortDesignatedRoot = $param->{'dot1dStpPortDesignatedRoot'}
		if ($param->{'dot1dStpPortDesignatedRoot'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortDesignatedRoot.$port"),
		ASN_OCTET_STR, DeHex($dot1dStpPortDesignatedRoot) );

	# dot1dStpPortDesignatedCost
	my $dot1dStpPortDesignatedCost = 100;
	$dot1dStpPortDesignatedCost = $param->{'dot1dStpPortDesignatedCost'}
		if ($param->{'dot1dStpPortDesignatedCost'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortDesignatedCost.$port"),
		ASN_INTEGER, $dot1dStpPortDesignatedCost);

	# dot1dStpPortDesignatedBridge
	# Default is '00:2A' followed by the MAC address of this bridge
	my $dot1dStpPortDesignatedBridge = '00:2A:' . $BridgeBaseMAC;
	$dot1dStpPortDesignatedBridge = $param->{'dot1dStpPortDesignatedBridge'}
		if ($param->{'dot1dStpPortDesignatedBridge'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortDesignatedBridge.$port"),
		ASN_OCTET_STR, DeHex($dot1dStpPortDesignatedBridge) );

	# dot1dStpPortDesignatedPort
	my $dot1dStpPortDesignatedPort = '80:01';
	$dot1dStpPortDesignatedPort = $param->{'dot1dStpPortDesignatedPort'}
		if ($param->{'dot1dStpPortDesignatedPort'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortDesignatedPort.$port"),
		ASN_OCTET_STR, DeHex($dot1dStpPortDesignatedPort) );

	# dot1dStpPortForwardTransitions
	my $dot1dStpPortForwardTransitions = 0;
	$dot1dStpPortForwardTransitions = $param->{'dot1dStpPortForwardTransitions'}
		if ($param->{'dot1dStpPortForwardTransitions'});
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dStpPortForwardTransitions.$port"),
		ASN_COUNTER, $dot1dStpPortForwardTransitions);

}


########################################################################
#

=item C<< $agent->AddTpFdbEntry($datahash) >>

Add a forwarding table entry

	$agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => $IFMAC,
                        dot1dTpFdbPort => $BridgePort,
                        dot1dTpFdbStatus => 4,
	} );

All three values MUST be specified.

Valid dot1dTpFdbStatus values are defined in RFC1493:

	  other(1),
	  invalid(2),
	  learned(3),
	  self(4),
	  mgmt(5)


=cut

#
########################################################################

sub AddTpFdbEntry ($$) {
	my ($self, $param) = @_;

	print "AddTpFdbEntry\n" if ($self->{debugging});

	# Sanity check the supplied values

	if (!$param->{'dot1dTpFdbAddress'}) {
		confess "AddTpFdbEntry called without a dot1dTpFdbAddress";
	}
	my $MAC = $param->{'dot1dTpFdbAddress'};
	if ($MAC !~ /^([0-9a-fA-F][0-9a-fA-F]:){5}[0-9a-fA-F][0-9a-fA-F]$/) {
		confess "AddTpFdbEntry called with bad dot1dTpFdbAddress";
	}

	if (!$param->{'dot1dTpFdbPort'}) {
		confess "AddTpFdbEntry called without a dot1dTpFdbPort";
	}
	my $port = $param->{'dot1dTpFdbPort'};

	if (!$param->{'dot1dTpFdbStatus'}) {
		confess "AddTpFdbEntry called without a dot1dTpFdbStatus";
	}
	my $status = $param->{'dot1dTpFdbStatus'};

	# Work out the OID component from the MAC address
	my $OIDpart = HexToOIDpart( $MAC );

	print "AddTpFdbEntry $MAC $port $status $OIDpart\n" if ($self->{debugging});

	# Build the Dot1dTpFdbEntry
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpFdbAddress.$OIDpart"),
		ASN_OCTET_STR, DeHex( $MAC ) );
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpFdbPort.$OIDpart"),
		ASN_INTEGER, $port );
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpFdbStatus.$OIDpart"),
		ASN_INTEGER, $status );

}


########################################################################
#

=item C<< $agent->AddTpPortEntry($datahash) >>

Construct a dot1dTpPortEntry.

        $agent->AddTpPortEntry( {
                        dot1dTpPort => $BridgePort,
                        dot1dTpPortMaxInfo => 1524,
                        dot1dTpPortInFrames => 0,
                        dot1dTpPortOutFrames => 0,
                        dot1dTpPortInDiscards => 0,
        } );

dot1dTpPort must be specified. The other parameters will default to the
values shown above if not set explicitly.

=cut
#
########################################################################

sub AddTpPortEntry ($$) {
	my ($self, $param) = @_;

	# Check for mandatory parameter
	if (!$param->{'dot1dTpPort'}) {
		confess "AddTpFdbEntry called without a dot1dTpPort";
	}
	my $dot1dTpPort = $param->{'dot1dTpPort'};

	print "AddTpPortEntry $dot1dTpPort\n" if ($self->{debugging});

	# Default values
	my $dot1dTpPortMaxInfo = 1524;
	my $dot1dTpPortInFrames = 0;
	my $dot1dTpPortOutFrames = 0;
	my $dot1dTpPortInDiscards = 0;

	# Handle remaining parameters
	if ($param->{'dot1dTpPortMaxInfo'}) {
		$dot1dTpPortMaxInfo = $param->{'dot1dTpPortMaxInfo'};
	}
	if ($param->{'dot1dTpPortInFrames'}) {
		$dot1dTpPortInFrames = $param->{'dot1dTpPortInFrames'};
	}
	if ($param->{'dot1dTpPortOutFrames'}) {
		$dot1dTpPortOutFrames = $param->{'dot1dTpPortOutFrames'};
	}
	if ($param->{'dot1dTpPortInDiscards'}) {
		$dot1dTpPortInDiscards = $param->{'dot1dTpPortInDiscards'};
	}

	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpPort.$dot1dTpPort"),
		ASN_INTEGER, $dot1dTpPort );
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpPortMaxInfo.$dot1dTpPort"),
		ASN_INTEGER, $dot1dTpPortMaxInfo );
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpPortInFrames.$dot1dTpPort"),
		ASN_COUNTER, $dot1dTpPortInFrames );
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpPortOutFrames.$dot1dTpPort"),
		ASN_COUNTER, $dot1dTpPortOutFrames );
	$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dTpPortInDiscards.$dot1dTpPort"),
		ASN_COUNTER, $dot1dTpPortInDiscards );
}


########################################################################
#

=item C<< $agent->SetBridge($datahash) >>

Sets the basic bridge parameters for a bridge instance

	$agent->SetBridge( {
		dot1dBaseBridgeAddress => '00:0C:4A:27:65:BF',
	} );

=cut

#
########################################################################

sub SetBridge ($$) {
	my ($self, $param) = @_;

	# dot1dBaseBridgeAddress
	if ($param->{'dot1dBaseBridgeAddress'}) {
		my $dot1dBaseBridgeAddress = $param->{'dot1dBaseBridgeAddress'};
		$self->SetMIBValue( CanonicalOID("BRIDGE-MIB::dot1dBaseBridgeAddress.0"),
			ASN_OCTET_STR, DeHex($dot1dBaseBridgeAddress) );
	}
}

########################################################################
#

=item C<< $agent->AddIpRouteEntry($datahash) >>

Add an entry to the routing table e.g.:

	$agent->AddIpRouteEntry ({
		ipRouteDest => '10.1.3.0',
		ipRouteIfIndex => 2,
		ipRouteMetric1 => -1,			*
		ipRouteMetric2 => -1,			*
		ipRouteMetric3 => -1,			*
		ipRouteMetric4 => -1,			*
		ipRouteNextHop => '10.0.2.34',
		ipRouteType => 4,			*
		ipRouteProto => 2,			*
		ipRouteAge => 0,			*
		ipRouteMask => '255.255.255.0',
		ipRouteMetric5 => -1,			*
		ipRouteInfo => '.0.0',			*
	});

Items marked '*' above can be omitted, and will default to the
values shown.

Valid values for ipRouteType are:

	1	Other
	2	Invalidated
	3	Direct (i.e. a directly connected network)
	4	Indirect (i.e. via a router)

=cut

#
########################################################################

sub AddIpRouteEntry ($$) {
	my ($self, $param) = @_;

	print "AddIpRouteEntry $param->{'ipRouteDest'}\n" if $self->{debugging};
	my $ipRouteDest = $param->{'ipRouteDest'};
	confess "AddIpRouteEntry called without ipRouteDest" if (!defined($ipRouteDest));
	AssertIP($ipRouteDest);
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteDest.$ipRouteDest"),
		ASN_IPADDRESS, IPTo4Bytes($ipRouteDest) );

	my $ipRouteNextHop = $param->{'ipRouteNextHop'};
	confess "AddIpRouteEntry called without ipRouteNextHop" if (!defined($ipRouteNextHop));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteNextHop.$ipRouteDest"),
		ASN_IPADDRESS, IPTo4Bytes($ipRouteNextHop) );

	my $ipRouteMask = $param->{'ipRouteMask'};
	confess "AddIpRouteEntry called without ipRouteMask" if (!defined($ipRouteMask));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteMask.$ipRouteDest"),
		ASN_IPADDRESS, IPTo4Bytes($ipRouteMask) );

	my $ipRouteIfIndex = $param->{'ipRouteIfIndex'};
	confess "AddIpRouteEntry called without ipRouteIfIndex" if (!defined($ipRouteIfIndex));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteIfIndex.$ipRouteDest"),
		ASN_INTEGER, $ipRouteIfIndex );

	my $ipRouteMetric1 = $param->{'ipRouteMetric1'};
	$ipRouteMetric1 = -1 if (!defined($ipRouteMetric1));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteMetric1.$ipRouteDest"),
		ASN_INTEGER, $ipRouteMetric1 );

	my $ipRouteMetric2 = $param->{'ipRouteMetric2'};
	$ipRouteMetric2 = -1 if (!defined($ipRouteMetric2));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteMetric2.$ipRouteDest"),
		ASN_INTEGER, $ipRouteMetric2 );

	my $ipRouteMetric3 = $param->{'ipRouteMetric3'};
	$ipRouteMetric3 = -1 if (!defined($ipRouteMetric3));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteMetric3.$ipRouteDest"),
		ASN_INTEGER, $ipRouteMetric3 );

	my $ipRouteMetric4 = $param->{'ipRouteMetric4'};
	$ipRouteMetric4 = -1 if (!defined($ipRouteMetric4));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteMetric4.$ipRouteDest"),
		ASN_INTEGER, $ipRouteMetric4 );

	my $ipRouteMetric5 = $param->{'ipRouteMetric5'};
	$ipRouteMetric5 = -1 if (!defined($ipRouteMetric5));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteMetric5.$ipRouteDest"),
		ASN_INTEGER, $ipRouteMetric5 );

	my $ipRouteType = $param->{'ipRouteType'};
	$ipRouteType = 4 if (!defined($ipRouteType));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteType.$ipRouteDest"),
		ASN_INTEGER, $ipRouteType );

	my $ipRouteProto = $param->{'ipRouteProto'};
	$ipRouteProto = 2 if (!defined($ipRouteProto));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteProto.$ipRouteDest"),
		ASN_INTEGER, $ipRouteProto );

	my $ipRouteAge = $param->{'ipRouteAge'};
	$ipRouteAge = 0 if (!defined($ipRouteAge));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteAge.$ipRouteDest"),
		ASN_INTEGER, $ipRouteAge );

	my $ipRouteInfo = $param->{'ipRouteInfo'};
	$ipRouteInfo = '.0.0' if (!defined($ipRouteInfo));
        $self->SetMIBValue( CanonicalOID("RFC1213-MIB::ipRouteInfo.$ipRouteDest"),
		ASN_OBJECT_ID, $ipRouteInfo );
}

######################################################################
######################################################################
# Runtime procedures
######################################################################
######################################################################

######################################################################
# Given an OID in internal form, find 'next' OID in MIB
# Return value is in string form so that it can index into mibdata
######################################################################
sub NextOID ($$) {
	my ($self, $oid) = @_;

	# First make sure that the oidlist is ready
	$self->BuildOIDList();

	# Convert the supplied OID to string form
	my $oidstr = '.';
	$oidstr .= join '.', $oid->to_array;
	print "NextOID($oidstr) -> " if ($self->{debugging});

	# If the supplied OID is in the MIB, we already know what comes next
	my $index = ${$self->{oidlistindex}}{$oidstr};
	if (defined($index) and defined(${$self->{oidlist}}[$index+1])) {
		my $retval = '.';
		$retval .= join '.', ${$self->{oidlist}}[$index+1]->to_array;

		my $iplus = $index + 1;
		print "oidlist[$iplus] -> $retval\n" if ($self->{debugging});
	}

	# Not in MIB, so scan through oidlist looking for
	# first OID greater than this one
	my $n;
	for ($n=0; defined(${$self->{oidlist}}[$n]); $n++) {
		if (${$self->{oidlist}}[$n] > $oid) {
			my $retval = '.';
			$retval .= join '.', ${$self->{oidlist}}[$n]->to_array;

			print "scan -> $retval\n" if ($self->{debugging});
			return $retval;
		}
	}

	# Off the end!
	print "Result undefined\n" if ($self->{debugging});
	return undef;
}

######################################################################
#
# define a subroutine to actually handle the incoming requests to our
# part of the OID tree.  This subroutine will get called for all
# requests within the OID space under the registration OID made
# in the new() method above.
#
######################################################################

sub my_snmp_handler () {
	my $self = shift;
	my ($handler, $registration_info, $request_info, $requests) = @_;
	my $request;

	print STDERR "refs: ",join(", ", ref($handler),
		ref($registration_info), 
		ref($request_info),
		ref($requests)),"\n" if ($self->{debugging} > 1);

	print STDERR "self is $self\n" if ($self->{debugging} > 1);
	print STDERR "registration_info is $registration_info\n" if ($self->{debugging} > 1);
	print STDERR "processing a request of type " . $request_info->getMode() . "\n" if ($self->{debugging});

	# print "Target: ", inet_ntoa($request_info->getTargetIp()), "\n";

	# Loop through the variables asked for in this request
	for($request = $requests; $request; $request = $request->next()) {
		my $oid = $request->getOID();
		print STDERR "  processing request for $oid\n" if ($self->{debugging});

		if ($request_info->getMode() == MODE_GET) {
			# if the requested oid is one that we know, then return the data
			my $oidstr = '.';
			$oidstr .= join '.', $oid->to_array;
			print "OID is $oidstr\n" if ($self->{debugging});

			if (defined($self->{mibdata}->{$oidstr})) {
				my $item = $self->{mibdata}->{$oidstr};

				my $value = $item->{'value'};
				# We may have a function or closure here. If so, invoke it.
				$value = $value->gen() if ref $value;
				my $length = length  $value;
				my $check = unpack 'H*', $value;
				print STDERR "   -> $value $check ($length)\n" if ($self->{debugging});

				$request->setValue($item->{'type'}, $value);
			}
			else {
				print STDERR "Request for undefined object: $oid\n" if ($self->{debugging});

				# Bodge in some test data
				# $request->setValue(ASN_OCTET_STR, (pack 'H12', '010203040506'));
				# $request->setValue(ASN_IPADDRESS, (pack 'H8', '01020304'));
				# $request->setValue(ASN_OBJECT_ID, ".1.1.6.1.2.1.1.4.0");
				# $request->setValue(ASN_OBJECT_ID, ".1.3.6.1.2.1.1.5.0" );
			}
		} elsif ($request_info->getMode() == MODE_GETNEXT) {
			# We must return OID and value for the next item after the OID supplied
			my $oidstr = '.';
			$oidstr .= join '.', $oid->to_array;
			print "OID is $oidstr\n" if ($self->{debugging});

			my $nextoid = $self->NextOID($oid);

			if (defined($nextoid) and defined($self->{mibdata}->{$nextoid})) {
				my $item = $self->{mibdata}->{$nextoid};

				my $value = $item->{'value'};
				# We may have a function or closure here. If so, invoke it.
				$value = $value->gen() if ref $value;
				my $length = length  $value;
				my $check = unpack 'H*', $value;
				print STDERR "   -> $value $check ($length)\n" if ($self->{debugging});

				$request->setOID($nextoid);
				$request->setValue($item->{'type'}, $value);
			}
		} elsif ($request_info->getMode() == MODE_SET_RESERVE1) {
			# Do error checking here so that multi-value SETs will succeed or fail as a unit.

			# At the moment we only accept SET operations for MIB items that are already defined
			my $oidstr = '.';
			$oidstr .= join '.', $oid->to_array;
			print "OID is $oidstr\n" if ($self->{debugging});

			my $item = $self->{mibdata}->{$oidstr};
			if (not defined($item)) {
				$request->setError($request_info, SNMP_ERR_NOSUCHNAME);
			}
			# We may have a function or closure here. If so, we do not want to over-write it.
			my $value = $item->{'value'};
			if (ref $value) {
				$request->setError($request_info, SNMP_ERR_NOTWRITABLE);
			}
		} elsif ($request_info->getMode() == MODE_SET_ACTION) {
			# if the requested oid is one that we know, then store the data
			my $oidstr = '.';
			$oidstr .= join '.', $oid->to_array;
			print "OID is $oidstr\n" if ($self->{debugging});

			# Sanity check - this should have been tested in the pre-SET phase
			if (not defined($self->{mibdata}->{$oidstr})) {
				$request->setError($request_info, SNMP_ERR_NOSUCHNAME);
				return undef;
			}

			# Find the current item and value
			my $item = $self->{mibdata}->{$oidstr};
			my $value = $item->{'value'};

			# Extract the proposed new value from the SET request
			# (Note that we cannot extract the type :-()
			my $newvalue = $request->getValue();

			my $length = length  $newvalue;
			my $check = unpack 'H*', $newvalue;
			print STDERR "SET   -> $newvalue $check ($length)\n" if ($self->{debugging});

			# Store the new value
			$item->{'value'} = $newvalue;

			# Return the new value
			$request->setValue($item->{'type'}, $newvalue);
		}
	}

	print STDERR "  finished processing\n"
	if ($self->{debugging});
}




######################################################################
######################################################################
#

=back

=head1 Examples

For detailed code examples, see the C<examples> directory in the
Raddle distribution.

=head1 Distribution

Raddle is distributed from SourceForge: http://raddle.sourceforge.net/

=head1 Author

Andrew Findlay <andrew.findlay@skills-1st.co.uk>

=cut

#
######################################################################
######################################################################


######################################################################
# Module must return true
######################################################################
1;
######################################################################

