#!/usr/bin/perl -w
#
# TicksSince.pm
#
# Class to return number of timeticks since a given date/time
# To be tied to MIB items like SysUptime.0
#
# Andrew Findlay
# 10 April 2003
#
# $Id: TicksSince.pm,v 1.3 2009/11/13 13:29:55 afindlay Exp $

#############################################################################
# 
# Copyright (c) 2003 Andrew Findlay
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
#############################################################################


package Net::Raddle::TicksSince;

use strict;
use warnings;
use Carp;

our $VERSION = 0.07;

=head1 Net::Raddle::TicksSince

Variables tied to this class specify a boot-time in seconds
since the epoch. When read back later, the variable returns
the number of timeticks since boot. A timetick is 0.01s.
The result is returned to a precision of 100 timeticks (1s).

e.g.:

	use Net::Raddle::TicksSince;
	use Date::Calc qw(Mktime);

	my $SinceBoot;
	tie $SinceBoot, "Net::Raddle::TicksSince", Mktime( 2002, 9, 24, 3, 59, 14)
		or die "cannot tie the knot";
	print "It is now $SinceBoot ticks since boot-time\n";

TicksSince will throw an exception using I<die> if the boot-time is
not passed as an integer.

When used with Net::Raddle::SNMPAgent it is not necessary to include
the 'use' line. Statements llike this will work:

        $agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'),
                ASN_TIMETICKS, "Net::Raddle::TicksSince",
                Mktime( 2003, 1, 11, 8, 59, 14) );

Any request for sysUpTime.0 will now return the time since 08:59:14
on 11th January 2003, in unites of 0.01s. As this is a 32-bit quantity,
it wraps around after 497 days and a few hours.

=cut

sub TIESCALAR {
	my ($class, $boottime) = @_;

	confess "Net::Raddle::TicksSince needs an integer number of seconds since the epoch"
	if ($boottime !~ /^[0-9]+$/);

	return bless \$boottime, $class;
}

# Allow the boot time to be updated
sub STORE {
	my ($self, $value) = @_;

	confess "Net::Raddle::TicksSince needs an integer number of seconds since the epoch"
	if ($value !~ /^[0-9]+$/);

	$$self = $value;
}

sub FETCH {
	my ($self) = @_;

	return (time - $$self) * 100;
}

1;

