#!/usr/bin/perl -w
#
# TrafficGen.pm
#
# Class to return number of bytes or packets deemed to have passed since it was
# last called.
#
# Andrew Findlay
# 10 Nov 2009
#
# $Id: TrafficGen.pm,v 1.2 2009/11/13 13:29:55 afindlay Exp $

#############################################################################
# 
# Copyright (c) 2009 Andrew Findlay
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
#############################################################################


package Net::Raddle::TrafficGen;

use strict;
use warnings;
use Carp;

our $VERSION = 0.07;

=head1 Net::Raddle::TrafficGen

Objects of this class specify traffic parameters.  When the 'gen'
method is invoked, returns a number dependent on the time since the
last invocation and on the defined parameters.  The number increases
constantly to represent an SNMP counter.

e.g.:

	use Net::Raddle::TrafficGen;

	my $IFBytesIN = Net::Raddle::TrafficGen->new( {rate => 100000/8} )
		or die "cannot create traffic generator";
	print "We have seen ", $IFBytesIN->gen(), " bytes since boot\n";
	sleep 5;
	print "We have seen ", $IFBytesIN->gen(), " bytes since boot\n";
	sleep 5;
	print "We have seen ", $IFBytesIN->gen(), " bytes since boot\n";


TrafficGen will throw an exception using I<die> if the parameters
are not understood.

When used with Net::Raddle::SNMPAgent statements like this will work:

        $agent->TieMIBItem( CanonicalOID('IF-MIB::ifInOctets.102'),
                ASN_COUNTER, Net::Raddle::TrafficGen->new( {rate => 123000000/8, limit => 155000000/8} );

Any request for IF-MIB::ifInOctets.102 will now return a number of bytes
that increases in a semi-random manner, with a maximum bandwidth of 155Mbit/s

=head2 new()

The new() method creates an object of class Net::Raddle::TrafficGen and sets config parameters.

The parameters are passed as a pointer to a hash, whose keys may include:

=over

=item B<rate>
The mean rate to generate (units/second).
Defaults to limit/2 if not specified.

=item B<limit>
The maximum rate to generate (units/second).
Defaults to rate*2 if not specified.
At least one of B<rate> and B<limit> MUST be specified.

=item B<initial>
The initial value of the counter

=item B<mode>
The traffic generation mode: either "random" or "constant"

=back

=cut

sub new {
	my ($class, $config) = @_;

	# Extract config parameters
	#
	# Target rate (units/sec)
	my $rate = $config->{rate};
	# Limiting rate (units/sec)
	my $limit = $config->{limit};
	# Traffic generation mode
	my $mode = $config->{mode};
	# Initial value
	my $initial = $config->{initial};

	confess "Net::Raddle::TrafficGen requires either 'rate' or 'limit' to be specified"
	if ((not defined $rate) and (not defined $limit));

	if (not defined $rate) {
		$rate = $limit / 2;
	}

	if (not defined $limit) {
		$limit = $rate * 2;
	}

	if (not defined $mode) {
		$mode = "random";
	}

	confess "Net::Raddle::TrafficGen does not understand $mode as a traffic-generation mode"
	unless (($mode eq "constant") or ($mode eq "random"));

	# We need to store a few things to implement this
	my %self;

	# The per-second rate on this quantity
	$self{rate} = $rate;
	# The per-second limit on this quantity
	$self{limit} = $limit;
	# When the value was last read
	$self{lastcall} = time;
	# The value when it was last read
	$self{counter} = 0;
	$self{counter} = $initial if defined $initial;
	# The traffic-generation mode
	$self{mode} = $mode;

	# How many units are in the queue?
	$self{backlog} = 0;

	return bless \%self, $class;
}

=head2 gen()

Calculates a traffic increment and returns the current state of the counter

=cut

sub gen {
	my ($self) = @_;

	confess "Not a class method" unless ref $self;

	# How long since we were last called?
	my $timenow = time;
	my $delta = $timenow - $self->{lastcall};
	# Update last-call time
	$self->{lastcall} = $timenow;

	# Calculate a proposed increment to the counter
	my $inc = $self->{backlog};
	if ($self->{mode} eq "constant") {
		$inc += $delta * $self->{rate};
	}
	elsif ($self->{mode} eq "random") {
		$inc += int($delta * $self->{rate} * rand 2);
	}

	# Calculate the maximum that can be allowed in this period
	my $max =  $delta * $self->{limit};
	# Anything above the limit gets added to the backlog
	# (Note that we assume unlimited buffer space here)
	if ($inc > $max) {
		$self->{backlog} = $inc - $max;
		$inc = $max;
	}
	else {
		$self->{backlog} = 0;
	}

	$self->{counter} += $inc;

	# Modulo 32 bits. This may not be portable...
	$self->{counter} %= 2**32;

	return $self->{counter};
}

1;

