#!/usr/bin/perl -w
#
# r1.pl
# Definition of main site router

use lib '/usr/local/lib/raddle/branch-network';

use strict;

use Net::Raddle::SNMPAgent qw(IPToMAC Netmask CanonicalOID);
use Net::Raddle::TrafficGen;
use netconfig;
# We need the type definitions when adding individual values
use NetSNMP::ASN (':all');

# net-snmp cannot give us command-line parameters so we use an environment variable
# to set the debug level
#
my $debug = $ENV{RADDLE_DEBUG};
$debug = 0 if not $debug;
print "DEBUG: $debug\n" if $debug;

# Create the SNMP agent
my $agent = Net::Raddle::SNMPAgent->new( debugging => $debug )
	or die "Cannot set up the agent for r1";

# Load the boilerplate
# Use the installed version if present, otherwise look in current directory
$agent->ParseDataFile( '/usr/local/lib/raddle/branch-network/r1.snmp' )
	or $agent->ParseDataFile( 'r1.snmp' )
	or die( "Cannot read the r1.snmp boilerplate file" );

# Override name and location
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysName.0'), ASN_OCTET_STR,
                        "group-$GroupNumber-r1.class.example.org");
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysLocation.0'), ASN_OCTET_STR,
                        "Virtual comms rack $GroupNumber");

# Date and time of boot
$agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'), ASN_TIMETICKS, "Net::Raddle::TicksSince",
                (time - 7*24*3600 +8273) );


# Add Loopback interface
#
# First set up some traffic generators
my $LoopPkts = Net::Raddle::TrafficGen->new({rate => 100});
my $LoopBytes = Net::Raddle::TrafficGen->new({rate => 100*250});
#
my $LoopbackIndex = $agent->AddIF( {
			ifDescr => 'Loopback0',
			ifType => 24,
			ifSpeed => 4294967295,
			ifInOctets => $LoopBytes,
			ifInUcastPkts => $LoopPkts,
			ifOutOctets => $LoopBytes,
			ifOutUcastPkts => $LoopPkts,
	} );

# Bind IP address to Loopback
$agent->AddIpAddrEntry( {
			ipAdEntAddr => '127.0.0.1',
			ipAdEntIfIndex => $LoopbackIndex,
			ipAdEntNetMask => '255.0.0.0',
              } );

# For r1 we assume that all traffic flows straight through between
# the E1 line and the local fast ethernet. The site consumes more than it sends.
# Simulate with traffic generators:
#
my $linerate = 2000000;
my $BytesToSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.5/8, limit=>$linerate/8});
my $PktsToSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.5/(8*1000)});
my $BytesFromSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.07/8, limit=>$linerate/8});
my $PktsFromSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.07/(8*1000)});


# Add Fast Ether
my $fe0Index = $agent->AddIF( {
			ifDescr => 'FastEthernet0/0',
			ifAlias => 'Main site network',
			ifType => 6,
			ifPhysAddress => $R1ClassNetMAC,
			ifSpeed => 100000000,
			ifInOctets => $BytesToSite,
			ifInUcastPkts => $PktsToSite,
			ifOutOctets => $BytesFromSite,
			ifOutUcastPkts => $PktsFromSite,
			ifInUnknownProtos => int rand 1000,
	} );

# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $R1ClassNetAddr,
			ipAdEntIfIndex => $fe0Index,
			ipAdEntNetMask => Netmask( $ClassNetMaskLen ),
              } );

# Add serial line to remote site
my $serial0Index = $agent->AddIF( {
			ifDescr => 'Serial1/0',
			ifAlias => 'E1 line to remote site',
			ifType => 23,
			ifSpeed => 2000000,
			ifInOctets => $BytesFromSite,
			ifInUcastPkts => $PktsFromSite,
			ifOutOctets => $BytesToSite,
			ifOutUcastPkts => $PktsToSite,
			ifInUnknownProtos => int rand 1000,
	} );

# Bind IP address to serial line
$agent->AddIpAddrEntry( {
	ipAdEntAddr => "$R1SerialNetAddr",
	ipAdEntIfIndex => $serial0Index,
	ipAdEntNetMask => Netmask( $SerialNetMaskLen ),
} );


# Now some routing table entries
# JC - changes to correct ipRouteType - 3=direct, 4=indirect
# JC - fyi - ipRouteProto - 1=other, 2=local, 4=ICMP redirect, 8=rip, 11=ciscoIgrp
$agent->AddIpRouteEntry ({
         ipRouteDest => '0.0.0.0',
         ipRouteIfIndex => $fe0Index,
         ipRouteNextHop => $R1DefaultRoute,
         ipRouteType => 4,
         ipRouteProto => 2,
         ipRouteMask => '0.0.0.0',
});

$agent->AddIpRouteEntry ({
         ipRouteDest => "$SiteSuperNet",
         ipRouteIfIndex => $serial0Index,
         ipRouteNextHop => "$R2SerialNetAddr",
         ipRouteType => 4,
         ipRouteProto => 11,
         ipRouteMask => Netmask( $SiteSuperNetMaskLen ) ,
});

$agent->AddIpRouteEntry ({
         ipRouteDest => "$SerialNetNet",
         ipRouteIfIndex => $serial0Index,
         ipRouteNextHop => "$R1SerialNetAddr",
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask( $SerialNetMaskLen ),
});

# ARP table entries
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe0Index,
        ipNetToMediaPhysAddress => $R1ClassNetMAC,
        ipNetToMediaNetAddress => $R1ClassNetAddr,
        ipNetToMediaType => 4,
} );

