#!/usr/bin/perl -w
#
# r2.pl
#
# Emulation spec for remote site access router

use lib '/usr/local/lib/raddle/branch-network';

use strict;

use Net::Raddle::SNMPAgent qw(IPToMAC Netmask CanonicalOID);
use Net::Raddle::TrafficGen;
use netconfig;
# We need the type definitions when adding individual values
use NetSNMP::ASN (':all');

# net-snmp cannot give us command-line parameters so we use an environment variable
# to set the debug level
#
my $debug = $ENV{RADDLE_DEBUG};
$debug = 0 if not $debug;
print "DEBUG: $debug\n" if $debug;

# Create the agent
my $agent = Net::Raddle::SNMPAgent->new( debugging => $debug )
	or die "Could not initialise emulator";

# Load the boilerplate
# Use the installed version if present, otherwise look in current directory
$agent->ParseDataFile( '/usr/local/lib/raddle/branch-network/r2.snmp' )
        or $agent->ParseDataFile( 'r2.snmp' )
        or die( "Cannot read the r2.snmp boilerplate file" );


# Override name and location
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysName.0'), ASN_OCTET_STR,
			"group-$GroupNumber-r2.class.example.org");
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysLocation.0'), ASN_OCTET_STR, 
			"Virtual site $GroupNumber");

# This router was booted up at 08:22:17 on 11/01/2003
$agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'), ASN_TIMETICKS, "Net::Raddle::TicksSince",
		(time - 7*24*3600 +8765) );

# Add Loopback interface
# First set up some traffic generators
my $LoopPkts = Net::Raddle::TrafficGen->new({rate => 100});
my $LoopBytes = Net::Raddle::TrafficGen->new({rate => 100*250});
#
my $LoopbackIndex = $agent->AddIF( {
			ifDescr => 'Loopback0',
			ifType => 24,
			ifSpeed => 4294967295,
                        ifInOctets => $LoopBytes,
                        ifInUcastPkts => $LoopPkts,
                        ifOutOctets => $LoopBytes,
                        ifOutUcastPkts => $LoopPkts,
	} );

# Bind IP address to Loopback
$agent->AddIpAddrEntry( {
			ipAdEntAddr => '127.0.0.1',
			ipAdEntIfIndex => $LoopbackIndex,
			ipAdEntNetMask => '255.0.0.0',
              } );

# For r2 we assume that all traffic flows straight through between
# the E1 line and fast ethernet 0. The site consumes more than it sends.
# Ideally the traffic here should match that seen in r1 but we don't currently
# have communication between the emulators.
#
# Simulate with traffic generators, assuming average 1000 byte packets:
#
my $linerate = 2000000;
my $BytesToSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.5/8, limit=>$linerate/8});
my $PktsToSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.5/(8*1000)});
my $BytesFromSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.07/8, limit=>$linerate/8});
my $PktsFromSite = Net::Raddle::TrafficGen->new({rate => $linerate*0.07/(8*1000)});

# Add Fast Ether 0
my $fe0Index = $agent->AddIF( {
			ifDescr => 'FastEthernet1/0',
			ifAlias => "Group $GroupNumber remote site network A",
			ifType => 6,
			ifPhysAddress => IPToMAC('00:30', $R2NetAAddr),
			ifSpeed => 100000000,
			ifInOctets => $BytesFromSite,
			ifInUcastPkts => $PktsFromSite,
			ifOutOctets => $BytesToSite,
			ifOutUcastPkts => $PktsToSite,
			ifInUnknownProtos => int rand 1000,
	} );

# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $R2NetAAddr,
			ipAdEntIfIndex => $fe0Index,
			ipAdEntNetMask => Netmask(28),
              } );

# Add Fast Ether 1
# Assume some traffic here, but lower levels than for Ether 0 and with smaller packets
my $fe1Index = $agent->AddIF( {
			ifDescr => 'FastEthernet1/1',
			ifAlias => "Group $GroupNumber remote site network B",
			ifType => 6,
			ifPhysAddress => IPToMAC('00:30', $R2NetBAddr),
			ifSpeed => 100000000,
			ifInOctets => Net::Raddle::TrafficGen->new({rate => 100000/8}),
			ifInUcastPkts => Net::Raddle::TrafficGen->new({rate => 100000/(8*400)}),
			ifOutOctets => Net::Raddle::TrafficGen->new({rate => 30000/8}),
			ifOutUcastPkts => Net::Raddle::TrafficGen->new({rate => 30000/(8*250)}),
			ifInUnknownProtos => int rand 1000,
	} );

# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $R2NetBAddr,
			ipAdEntIfIndex => $fe1Index,
			ipAdEntNetMask => Netmask(28),
              } );

# Add serial line to main site
my $serial0Index = $agent->AddIF( {
			ifDescr => 'Serial2/0',
			ifAlias => 'E1 line to central site',
			ifType => 23,
			ifSpeed => 2000000,
			ifOutOctets => $BytesFromSite,
			ifOutUcastPkts => $PktsFromSite,
			ifInOctets => $BytesToSite,
			ifInUcastPkts => $PktsToSite,
			ifInUnknownProtos => int rand 1000,
	} );

# Bind IP address to serial line
$agent->AddIpAddrEntry( {
	ipAdEntAddr => "$R2SerialNetAddr",
	ipAdEntIfIndex => $serial0Index,
	ipAdEntNetMask => Netmask(30),
} );


# Now some routing table entries

# Default route
# JC - changes for route type
$agent->AddIpRouteEntry ({
         ipRouteDest => '0.0.0.0',
         ipRouteIfIndex => $serial0Index,
         ipRouteNextHop => $R2DefaultRoute,
         ipRouteType => 4,
         ipRouteProto => 2,
         ipRouteMask => '0.0.0.0',
});

# Serial line network
$agent->AddIpRouteEntry ({
         ipRouteDest => "$SerialNetNet",
         ipRouteIfIndex => $serial0Index,
         ipRouteNextHop => "$R2SerialNetAddr",
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask(30),
});

# Site net A
$agent->AddIpRouteEntry ({
         ipRouteDest => $NetANet,
         ipRouteIfIndex => $fe0Index,
         ipRouteNextHop => $R2NetAAddr,
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask(28),
});

# Site net B
$agent->AddIpRouteEntry ({
         ipRouteDest => $NetBNet,
         ipRouteIfIndex => $fe1Index,
         ipRouteNextHop => $R2NetBAddr,
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask(28),
});

# Site net C goes via site net A
$agent->AddIpRouteEntry ({
         ipRouteDest => $NetCNet,
         ipRouteIfIndex => $fe0Index,
         ipRouteNextHop => $R3NetAAddr,
         ipRouteType => 4,
         ipRouteProto => 2,
         ipRouteMask => Netmask(28),
});


# ARP table entries

# Our own address on site net A
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe0Index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R2NetAAddr ),
        ipNetToMediaNetAddress => $R2NetAAddr,
        ipNetToMediaType => 4,
} );

# Our own address on site net B
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe1Index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R2NetBAddr ),
        ipNetToMediaNetAddress => $R2NetBAddr,
        ipNetToMediaType => 4,
} );

# Server on net A
$agent->AddIpNetToMediaEntry( {
	ipNetToMediaIfIndex => $fe0Index,
	ipNetToMediaPhysAddress => IPToMAC( '08:00', $HostA1Addr ),
	ipNetToMediaNetAddress => $HostA1Addr,
} );

# Clients on net B
$agent->AddIpNetToMediaEntry( {
	ipNetToMediaIfIndex => $fe1Index,
	ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostB1Addr ),
	ipNetToMediaNetAddress => $HostB1Addr,
} );

$agent->AddIpNetToMediaEntry( {
	ipNetToMediaIfIndex => $fe1Index,
	ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostB2Addr ),
	ipNetToMediaNetAddress => $HostB2Addr,
} );

# Switch on network B
$agent->AddIpNetToMediaEntry( {
	ipNetToMediaIfIndex => $fe1Index,
	ipNetToMediaPhysAddress => $SwitchS2BaseMAC,
	ipNetToMediaNetAddress => $SwitchS2Addr,
} );

