#!/usr/bin/perl -w
#
# r3.pl
#
# Emulation spec for R3 - remote site internal router

use lib '/usr/local/lib/raddle/branch-network';

use strict;

use Net::Raddle::SNMPAgent qw(IPToMAC Netmask CanonicalOID);
use Net::Raddle::TrafficGen;
use netconfig;
# We need the type definitions when adding individual values
use NetSNMP::ASN (':all');

# net-snmp cannot give us command-line parameters so we use an environment variable
# to set the debug level
#
my $debug = $ENV{RADDLE_DEBUG};
$debug = 0 if not $debug;
print "DEBUG: $debug\n" if $debug;

# Create the agent
my $agent = Net::Raddle::SNMPAgent->new( debugging => $debug )
	or die "Could not initialise agent";

# Load the boilerplate
# Use the installed version if present, otherwise look in current directory
$agent->ParseDataFile( '/usr/local/lib/raddle/branch-network/r3.snmp' )
        or $agent->ParseDataFile( 'r3.snmp' )
        or die( "Cannot read the r3.snmp boilerplate file" );


# Override name and location
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysName.0'), ASN_OCTET_STR,
			"group-$GroupNumber-r3.class.example.org");
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysLocation.0'), ASN_OCTET_STR, 
			"Virtual site $GroupNumber");

# This router was booted up at ...
$agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'), ASN_TIMETICKS, "Net::Raddle::TicksSince",
		(time - 7*24*3600 -2675) );

# Add Loopback interface
#
# First set up some traffic generators
my $LoopPkts = Net::Raddle::TrafficGen->new({rate => 100});
my $LoopBytes = Net::Raddle::TrafficGen->new({rate => 100*250});
#
my $LoopbackIndex = $agent->AddIF( {
			ifDescr => 'Loopback0',
			ifType => 24,
			ifSpeed => 4294967295,
                        ifInOctets => $LoopBytes,
                        ifInUcastPkts => $LoopPkts,
                        ifOutOctets => $LoopBytes,
                        ifOutUcastPkts => $LoopPkts,
	} );

# Bind IP address to Loopback
$agent->AddIpAddrEntry( {
			ipAdEntAddr => '127.0.0.1',
			ipAdEntIfIndex => $LoopbackIndex,
			ipAdEntNetMask => '255.0.0.0',
              } );


# For r3 we assume that most traffic flows straight through between
# FE0 (net A) and FE2 (net C).
# Simulate with traffic generators:
#
my $BytesAtoC = Net::Raddle::TrafficGen->new({rate => 60000000/8, limit => 100000000/8});
my $PktsAtoC = Net::Raddle::TrafficGen->new({rate => 60000000/(8*1000)});
my $BytesCtoA = Net::Raddle::TrafficGen->new({rate => 7000000/8, limit => 100000000/8});
my $PktsCtoA = Net::Raddle::TrafficGen->new({rate => 7000000/(8*350)});
#

# Add Fast Ether
my $fe0Index = $agent->AddIF( {
			ifDescr => 'FastEthernet1/0',
			ifAlias => "Group $GroupNumber remote site network A",
			ifType => 6,
			ifPhysAddress => IPToMAC('00:30', $R3NetAAddr),
			ifSpeed => 100000000,
			ifInOctets => $BytesAtoC,
			ifInUcastPkts => $PktsAtoC,
			ifOutOctets => $BytesCtoA,
			ifOutUcastPkts => $PktsCtoA,
			# Throw in some unknown protocols
			ifInUnknownProtos => Net::Raddle::TrafficGen->new( {rate => 2000} ),
	} );

# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $R3NetAAddr,
			ipAdEntIfIndex => $fe0Index,
			ipAdEntNetMask => Netmask(28),
              } );

# Add Fast Ether
my $fe1Index = $agent->AddIF( {
			ifDescr => 'FastEthernet1/1',
			ifAlias => "Group $GroupNumber remote site network B",
			ifType => 6,
			ifPhysAddress => IPToMAC('00:30', $R3NetBAddr),
			ifSpeed => 100000000,
			ifInOctets => Net::Raddle::TrafficGen->new({rate => 50000/8}),
			ifInUcastPkts => Net::Raddle::TrafficGen->new({rate => 50000/(8*1400)}),
			ifOutOctets => Net::Raddle::TrafficGen->new({rate => 50000/8}),
			ifOutUcastPkts => Net::Raddle::TrafficGen->new({rate => 30000/(8*750)}),
			ifInUnknownProtos => Net::Raddle::TrafficGen->new( {rate => 1000} ),
	} );

# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $R3NetBAddr,
			ipAdEntIfIndex => $fe1Index,
			ipAdEntNetMask => Netmask(28),
              } );

# Add Fast Ether
my $fe2Index = $agent->AddIF( {
			ifDescr => 'FastEthernet1/2',
			ifAlias => "Group $GroupNumber remote site network C",
			ifType => 6,
			ifPhysAddress => IPToMAC('00:30', $R3NetCAddr),
			ifSpeed => 100000000,
			ifInOctets => $BytesCtoA,
			ifInUcastPkts => $PktsCtoA,
			ifOutOctets => $BytesAtoC,
			ifOutUcastPkts => $PktsAtoC,
			ifInUnknownProtos => Net::Raddle::TrafficGen->new( {rate => 4000} ),
	} );

# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $R3NetCAddr,
			ipAdEntIfIndex => $fe2Index,
			ipAdEntNetMask => Netmask(28),
              } );


# Now some routing table entries

# Default route
# JC - change route types
$agent->AddIpRouteEntry ({
         ipRouteDest => '0.0.0.0',
         ipRouteIfIndex => $fe0Index,
         ipRouteNextHop => $R3DefaultRoute,
         ipRouteType => 4,
         ipRouteProto => 2,
         ipRouteMask => '0.0.0.0',
});

# Site net A
$agent->AddIpRouteEntry ({
         ipRouteDest => $NetANet,
         ipRouteIfIndex => $fe0Index,
         ipRouteNextHop => $R3NetAAddr,
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask($NetAMaskLen),
});

# Site net B
$agent->AddIpRouteEntry ({
         ipRouteDest => $NetBNet,
         ipRouteIfIndex => $fe1Index,
         ipRouteNextHop => $R3NetBAddr,
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask($NetBMaskLen),
});

# Site net C
$agent->AddIpRouteEntry ({
         ipRouteDest => $NetCNet,
         ipRouteIfIndex => $fe2Index,
         ipRouteNextHop => $R3NetCAddr,
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask($NetCMaskLen),
});


# ARP table entries

# Our own address on site net A
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe0Index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R3NetAAddr ),
        ipNetToMediaNetAddress => $R3NetAAddr,
        ipNetToMediaType => 4,
} );

# Our own address on site net B
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe1Index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R3NetBAddr ),
        ipNetToMediaNetAddress => $R3NetBAddr,
        ipNetToMediaType => 4,
} );

# Our own address on site net C
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe2Index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R3NetCAddr ),
        ipNetToMediaNetAddress => $R3NetCAddr,
        ipNetToMediaType => 4,
} );

# Router R2 on net A
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe0Index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R2NetBAddr ),
        ipNetToMediaNetAddress => $R2NetBAddr,
} );


# Server on net A
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe0Index,
        ipNetToMediaPhysAddress => IPToMAC( '08:00', $HostA1Addr ),
        ipNetToMediaNetAddress => $HostA1Addr,
} );

# Clients on net B
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe1Index,
        ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostB1Addr ),
        ipNetToMediaNetAddress => $HostB1Addr,
} );

$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe1Index,
        ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostB2Addr ),
        ipNetToMediaNetAddress => $HostB2Addr,
} );

# Clients on net C
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe2Index,
        ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostC1Addr ),
        ipNetToMediaNetAddress => $HostC1Addr,
} );

$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe2Index,
        ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostC2Addr ),
        ipNetToMediaNetAddress => $HostC2Addr,
} );

$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe2Index,
        ipNetToMediaPhysAddress => IPToMAC( '08:4c', $HostC3Addr ),
        ipNetToMediaNetAddress => $HostC3Addr,
} );

$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $fe2Index,
        ipNetToMediaPhysAddress => $SwitchS1BaseMAC,
        ipNetToMediaNetAddress => $SwitchS1Addr,
} );

