#!/usr/bin/perl -w
#
# s1.pl
# Definition of an etherswitch - this one is based on a Cisco Catalyst 1900

use lib '/usr/local/lib/raddle/branch-network';

use strict;

use Net::Raddle::SNMPAgent qw( IPToMAC Netmask CanonicalOID);
use Net::Raddle::TrafficGen;
use netconfig;
# We need the type definitions when adding individual values
use NetSNMP::ASN (':all');

# net-snmp cannot give us command-line parameters so we use an environment variable
# to set the debug level
#
my $debug = $ENV{RADDLE_DEBUG};
$debug = 0 if not $debug;
print "DEBUG: $debug\n" if $debug;

# Create the agent
my $agent = Net::Raddle::SNMPAgent->new( debugging => $debug )
         or die "Could not initialise Net::Raddle::SNMPAgent";

# Load the boilerplate
# Use the installed version if present, otherwise look in current directory
$agent->ParseDataFile( '/usr/local/lib/raddle/branch-network/s1.snmp' )
        or $agent->ParseDataFile( 's1.snmp' )
        or die( "Cannot read the s1.snmp boilerplate file" );

# Override name and location
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysName.0'), ASN_OCTET_STR,
                        "group-$GroupNumber-s1.class.example.org");
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysLocation.0'), ASN_OCTET_STR,
                        "Virtual machine room $GroupNumber");

# Date and time of boot
$agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'), ASN_TIMETICKS, "Net::Raddle::TicksSince",
                (time - 7*24*3600 -876) );


# Set the basic bridge params
$agent->SetBridge({
			dot1dBaseBridgeAddress => $SwitchS1BaseMAC,
	});

# Set the basic spanning tree params
# This is the root bridge
$agent->SetSpanningTree({
			dot1dStpDesignatedRoot => $SwitchS1BaseMAC,
	});

# Array of port indices
my @portIndex;

# Create 24 standard interfaces and some odds
my $count;
for ($count=1; $count<=29; $count++) {
	my $ifName = "$count";
	$ifName = "AUI" if ($count == 25);
	$ifName = "A" if ($count == 26);
	$ifName = "B" if ($count == 27);
	$ifName = "PortChannel" if ($count == 28);
	$ifName = "CPU" if ($count == 29);

	my $ifSpeed = 10000000;
	$ifSpeed = 100000000 if (($count == 26) or ($count == 27));
	$ifSpeed = 0 if ($count == 28);

	# Work out last byte of interface MAC address and insert it
	my ($MACStem, $MACByte) = ($SwitchS1BaseMAC =~ /(.*:)([0-9A-Fa-f]{1,2})$/);
	my $MACHex = sprintf "%02X", (hex($MACByte) + $count) % 256;
	my $IFMAC = $MACStem . $MACHex;
	if ($count == 29) {
		# CPU port uses the base MAC address
		$IFMAC = $SwitchS1BaseMAC;
	}
	# print "$count    $IFMAC\n";

	# Set up some traffic generators
	my $howbusy = int( rand $ifSpeed );
	# CPU port never very busy
	$howbusy /= 50 if ($count == 29);
	#
	my $InBytes = Net::Raddle::TrafficGen->new( {rate => $howbusy/8, limit => $ifSpeed/8} );
	my $InPkts = Net::Raddle::TrafficGen->new( {rate => $howbusy/(8*350)} );
	my $InUnknown = Net::Raddle::TrafficGen->new( {rate => (rand 0.10) * $howbusy/(8*350)} );
	my $OutBytes = Net::Raddle::TrafficGen->new( {rate => $howbusy/8, limit => $ifSpeed/8} );
	my $OutPkts = Net::Raddle::TrafficGen->new( {rate => $howbusy/(8*1200)} );

	my $NextIndex = $agent->AddIF( {
                        ifDescr => $ifName,
                        ifName => $ifName,
                        ifType => 6,
                        ifPhysAddress => $IFMAC,
                        ifSpeed => $ifSpeed,
                        ifInOctets => $InBytes,
                        ifInUcastPkts => $InPkts,
                        ifOutOctets => $OutBytes,
                        ifOutUcastPkts => $OutPkts,
                        ifInUnknownProtos => $InUnknown,
        } );

	# Stash the index for use later
	$portIndex[$count] = $NextIndex;

	# Add bridge port to each interface
	my $BridgePort = $agent->AddBridgePort( {
			dot1dBasePortIfIndex => $NextIndex,
	} );

	# Add Spanning Tree port data to each interface
	$agent->AddStpPortEntry( {
			dot1dStpPort => $BridgePort,
	} );

	# Add a Forwarding Database entry for the port's own MAC
	# Status 4 indicates 'self'
	$agent->AddTpFdbEntry( {
			dot1dTpFdbAddress => $IFMAC,
			dot1dTpFdbPort => $BridgePort,
			dot1dTpFdbStatus => 4,
	} );

	# Add a dot1dTpPortEntry for each port
	$agent->AddTpPortEntry( {
			dot1dTpPort => $BridgePort,
			dot1dTpPortMaxInfo => 1524,
			dot1dTpPortInFrames => $InPkts,
			dot1dTpPortOutFrames => $OutPkts,
			dot1dTpPortInDiscards => 0,
	} );
}


# This switch always has its IP management station on port 'CPU'
# which we created as portIndex 29
#

# Bind IP address to CPU port
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $SwitchS1Addr,
			ipAdEntIfIndex => $portIndex[29],
			ipAdEntNetMask => Netmask( $NetCMaskLen ),
              } );

# Fill in the ARP table
#
# Our own address (port 29)
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => 29,
        ipNetToMediaPhysAddress => $SwitchS1BaseMAC,
        ipNetToMediaNetAddress => $SwitchS1Addr,
        ipNetToMediaType => 4,
} );

# Router R3 on net C
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => 29,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R3NetCAddr ),
        ipNetToMediaNetAddress => $R3NetCAddr,
} );




# Now some routing table entries
# JC - change route types
$agent->AddIpRouteEntry ({
         ipRouteDest => '0.0.0.0',
         ipRouteIfIndex => $portIndex[29],
         ipRouteNextHop => $R3NetCAddr,
         ipRouteType => 4,
         ipRouteProto => 2,
         ipRouteMask => '0.0.0.0',
});

$agent->AddIpRouteEntry ({
         ipRouteDest => "$NetCNet",
         ipRouteIfIndex => $portIndex[29],
         ipRouteNextHop => "$SwitchS1Addr",
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask( $NetCMaskLen ) ,
});

# Add an ARP table
# This is our own address
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => 29,
        ipNetToMediaPhysAddress => $SwitchS1BaseMAC,
        ipNetToMediaNetAddress => $SwitchS1Addr,
        ipNetToMediaType => 4,
} );

# This one for R3
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => 29,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R3NetCAddr ),
        ipNetToMediaNetAddress => $R3NetCAddr,
} );


# Now connect some hosts to the ports
# This switch serves net C

        # Port A R3
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '00:30', $R3NetCAddr ),
                        dot1dTpFdbPort => 26,
                        dot1dTpFdbStatus => 3,
        } );

        # Port 2, Host C1
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '08:4c', $HostC1Addr ),
                        dot1dTpFdbPort => 1,
                        dot1dTpFdbStatus => 3,
        } );

        # Port 3, Host C2
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '08:4c', $HostC2Addr ),
                        dot1dTpFdbPort => 2,
                        dot1dTpFdbStatus => 3,
        } );

        # Port 4, Host C3
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '08:4c', $HostC3Addr ),
                        dot1dTpFdbPort => 3,
                        dot1dTpFdbStatus => 3,
        } );


