#!/usr/bin/perl -w
#
# s2.pl
# Definition of an etherswitch

use lib '/usr/local/lib/raddle/branch-network';

use strict;

use Net::Raddle::SNMPAgent qw( IPToMAC Netmask CanonicalOID);
use Net::Raddle::TrafficGen;
use netconfig;
# We need the type definitions when adding individual values
use NetSNMP::ASN (':all');

# net-snmp cannot give us command-line parameters so we use an environment variable
# to set the debug level
#
my $debug = $ENV{RADDLE_DEBUG};
$debug = 0 if not $debug;
print "DEBUG: $debug\n" if $debug;

# Create the agent
my $agent = Net::Raddle::SNMPAgent->new( debugging => $debug )
         or die "Could not initialise Net::Raddle::SNMPAgent";

# Load the boilerplate
# Use the installed version if present, otherwise look in current directory
$agent->ParseDataFile( '/usr/local/lib/raddle/branch-network/s2.snmp' )
        or $agent->ParseDataFile( 's2.snmp' )
        or die( "Cannot read the s2.snmp boilerplate file" );

# Override name and location
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysName.0'), ASN_OCTET_STR,
                        "group-$GroupNumber-s2.class.example.org");
$agent->SetMIBValue( CanonicalOID('SNMPv2-MIB::sysLocation.0'), ASN_OCTET_STR,
                        "Virtual machine room $GroupNumber");

# Date and time of boot
$agent->TieMIBItem( CanonicalOID('SNMPv2-MIB::sysUpTime.0'), ASN_TIMETICKS, "Net::Raddle::TicksSince",
                (time - 7*24*3600 - 17000) );


# Set the basic bridge params
$agent->SetBridge({
			dot1dBaseBridgeAddress => $SwitchS2BaseMAC,
	});

# Set the basic spanning tree params
# This is the root bridge
$agent->SetSpanningTree({
			dot1dStpDesignatedRoot => $SwitchS2BaseMAC,
	});

# This switch always has its IP management station on VLAN1
# which is interface index 1
#
my  $VLAN1index = $agent->AddIF( {
                        ifDescr => 'VLAN1',
                        ifName => 'VL1',
                        ifType => 6,
                        ifPhysAddress => $SwitchS2BaseMAC,
                        ifSpeed => 100000000,
                        ifInOctets => Net::Raddle::TrafficGen->new( {rate => 50000/8} ),
                        ifInUcastPkts => Net::Raddle::TrafficGen->new( {rate => 50000/(8*160)} ),
                        ifOutOctets => Net::Raddle::TrafficGen->new( {rate => 50000/8} ),
                        ifOutUcastPkts => Net::Raddle::TrafficGen->new( {rate => 50000/(8*500)} ),
                        ifInUnknownProtos => Net::Raddle::TrafficGen->new( {rate => 20} ),
        } );


# Bind IP address to Fast Ether
$agent->AddIpAddrEntry( {
			ipAdEntAddr => $SwitchS2Addr,
			ipAdEntIfIndex => $VLAN1index,
			ipAdEntNetMask => Netmask( $NetBMaskLen ),
              } );

# Create 24 standard interfaces
my $count;
for ($count=1; $count<=24; $count++) {
	# Work out last byte of interface MAC address and insert it
	my ($MACStem, $MACByte) = ($SwitchS2BaseMAC =~ /(.*:)([0-9A-Fa-f]{1,2})$/);
	my $MACHex = sprintf "%02X", (hex($MACByte) + $count) % 256;
	my $IFMAC = $MACStem . $MACHex;
	# print "$count    $IFMAC\n";

	# Set up some traffic generators
	my $ifSpeed = 100000000;
	my $howbusy = int( rand $ifSpeed );
	#
        my $InBytes = Net::Raddle::TrafficGen->new( {rate => $howbusy/8, limit => $ifSpeed/8} );
        my $InPkts = Net::Raddle::TrafficGen->new( {rate => $howbusy/(8*350)} );
        my $InUnknown = Net::Raddle::TrafficGen->new( {rate => (rand 0.10) * $howbusy/(8*350)} );
        my $OutBytes = Net::Raddle::TrafficGen->new( {rate => $howbusy/8, limit => $ifSpeed/8} );
        my $OutPkts = Net::Raddle::TrafficGen->new( {rate => $howbusy/(8*1200)} );

	my $NextIndex = $agent->AddIF( {
                        ifDescr => "FastEthernet0/$count",
                        ifName => "Fa0/$count",
                        ifType => 6,
                        ifPhysAddress => $IFMAC,
                        ifSpeed => $ifSpeed,
                        ifInOctets => $InBytes,
                        ifInUcastPkts => $InPkts,
                        ifOutOctets => $OutBytes,
                        ifOutUcastPkts => $OutPkts,
                        ifInUnknownProtos => $InUnknown,
        } );

	# Add bridge port to each interface
	my $BridgePort = $agent->AddBridgePort( {
			dot1dBasePortIfIndex => $NextIndex,
	} );

	# Add Spanning Tree port data to each interface
	$agent->AddStpPortEntry( {
			dot1dStpPort => $BridgePort,
	} );

	# Add a Forwarding Database entry for the port's own MAC
	# Status 4 indicates 'self'
	$agent->AddTpFdbEntry( {
			dot1dTpFdbAddress => $IFMAC,
			dot1dTpFdbPort => $BridgePort,
			dot1dTpFdbStatus => 4,
	} );

	# Add a dot1dTpPortEntry for each port
	$agent->AddTpPortEntry( {
			dot1dTpPort => $BridgePort,
			dot1dTpPortMaxInfo => 1524,
			dot1dTpPortInFrames => $InPkts,
			dot1dTpPortOutFrames => $OutPkts,
			dot1dTpPortInDiscards => 0,
	} );
}


# Add Null interface
my $NullIndex = $agent->AddIF( {
			ifDescr => 'NULL0',
			ifName => 'Nu0',
			ifType => 1,
			ifSpeed => 4294967295,
	} );

# Now some routing table entries
# JC - change route types
$agent->AddIpRouteEntry ({
         ipRouteDest => '0.0.0.0',
         ipRouteIfIndex => $VLAN1index,
         ipRouteNextHop => $R2NetBAddr,
         ipRouteType => 4,
         ipRouteProto => 2,
         ipRouteMask => '0.0.0.0',
});

$agent->AddIpRouteEntry ({
         ipRouteDest => "$NetBNet",
         ipRouteIfIndex => $VLAN1index,
         ipRouteNextHop => "$SwitchS2Addr",
         ipRouteType => 3,
         ipRouteProto => 2,
         ipRouteMask => Netmask( $NetBMaskLen ) ,
});

# Add ARP table entry for this switch
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $VLAN1index,
        ipNetToMediaPhysAddress => $SwitchS2BaseMAC,
        ipNetToMediaNetAddress => $SwitchS2Addr,
        ipNetToMediaType => 4,
} );

# Add ARP table entry for router R2
$agent->AddIpNetToMediaEntry( {
        ipNetToMediaIfIndex => $VLAN1index,
        ipNetToMediaPhysAddress => IPToMAC( '00:30', $R2NetBAddr ),
        ipNetToMediaNetAddress => $R2NetBAddr,
} );

# Now connect some hosts to the ports
# This swich serves net B

        # Port 1 (numbered 2), R2
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '00:30', $R2NetBAddr ),
                        dot1dTpFdbPort => 2,
                        dot1dTpFdbStatus => 3,
        } );

        # Port 2, Host B1
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '08:4c', $HostB1Addr ),
                        dot1dTpFdbPort => 3,
                        dot1dTpFdbStatus => 3,
        } );

        # Port 3, Host B2
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '08:4c', $HostB2Addr ),
                        dot1dTpFdbPort => 4,
                        dot1dTpFdbStatus => 3,
        } );


        # Port 24, R3
        $agent->AddTpFdbEntry( {
                        dot1dTpFdbAddress => IPToMAC( '00:30', $R3NetBAddr ),
                        dot1dTpFdbPort => 24,
                        dot1dTpFdbStatus => 3,
        } );

